package de.desy.web2cToolkit.web2c.web2cComponents;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;

import de.desy.web2cToolkit.web2c.web2cChatUtilities.Web2cChatChannel;
import de.desy.web2cToolkit.web2c.web2cDeviceAccess.Web2cCommonDeviceAccess;
import de.desy.web2cToolkit.web2c.web2cVideoUtilities.Web2cCommonVideoConnector;
import de.desy.web2cToolkit.web2c.web2cComponents.Web2cSection;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cFileLogger;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cKeyBoardListener;
import de.desy.web2cToolkit.web2cUtil.web2cFileIO.Web2cConfigReader;

/**
 * Embedded Page
 * @author R.Bacher, DESY
 */
public class Web2cEmbeddedPage {
	
	private ArrayList<ArrayList<String>> sectionsHTML = new ArrayList<ArrayList<String>>();
	private ArrayList<ArrayList<String>> sectionsJavaScript = new ArrayList<ArrayList<String>>();	
	private ArrayList<Web2cSection> sections = new ArrayList<Web2cSection>();
	private ArrayList<Web2cEmbeddedPage> embeddedPages = new ArrayList<Web2cEmbeddedPage>();
	private String homeDir = null;
	private int language = 0;
	private double xScale = 0.0;
	private double yScale = 0.0;
	private String referencePoint = null;
	private String referenceId = null;
	private String userName = null;
	private String userPrivilege = null;
	private Web2cCommonDeviceAccess commonDeviceAccess = null;
	private Web2cCommonVideoConnector commonVideoConnector = null;
	private Web2cChatChannel chatChannel = null;
	private Web2cFileLogger debugLogger = null;
	private Web2cKeyBoardListener debugListener = null;
	private String web2cSession = null;
	private String pageConfigFile = null;
	
	private boolean simulation = true;
	private String error = null;
	private String accessMode = "";
	private volatile Double[] constructionProgress = new Double[1];
	
	/**
	 * Embedded Page
	 * @param session Web2c session id
	 * @param home directory which contains user-specific files
	 * @param configFile configuration file
	 * @param name actual user name
	 * @param lang language
	 * @param refId ID of embedded page
	 * @param refPoint coordinates (left upper corner) of embedded page
	 * @param x horizontal scaling factor
	 * @param y vertical scaling factor
	 * @param privilege actual user privilege
	 * @param simulatedData flag for simulated or real data
	 * @param cda actual common device access instance
	 * @param cvc actual common video connector
	 * @param cc actual chat channel instance
	 * @param logger debug logger
	 * @param listener debug flag listener
	 * @param access access mode
	 * @param progress construction progress
	 */
	public Web2cEmbeddedPage(String session, String home, String configFile, String name, int lang, String refId, String refPoint, double x, double y, String privilege, boolean simulatedData, Web2cCommonDeviceAccess cda, Web2cCommonVideoConnector cvc, Web2cChatChannel cc, Web2cFileLogger logger, Web2cKeyBoardListener listener, String access, Double[] progress) {

		web2cSession = session;
		homeDir = home;
		userName = name;
		language = lang;
		xScale = x;
		yScale = y;
		referenceId = refId;
		referencePoint = refPoint;
		userPrivilege = privilege;
		simulation = simulatedData;
		commonDeviceAccess = cda;
		commonVideoConnector = cvc;
		chatChannel = cc;
		debugLogger = logger;
		debugListener = listener;
		pageConfigFile = configFile;
		accessMode = access;
		constructionProgress = progress;
		constructPage();
	}
	
	/**
	 * Returns full contents of page (HTML tags, JavaScript snippets) 
	 * @return list
	 */
	public ArrayList<Web2cComponent> getComponents() {
		
		ArrayList<Web2cComponent> list = new ArrayList<Web2cComponent>();
		for (Iterator<Web2cSection> it1 = sections.iterator(); it1.hasNext();) {
			Web2cSection section = (Web2cSection)it1.next();
			list.add((Web2cComponent)section);
			for (Iterator<Web2cComponent> it2 = (section.getComponents()).iterator(); it2.hasNext();) {
				list.add(it2.next());
			}
		}
		for (Iterator<Web2cEmbeddedPage> it1 = embeddedPages.iterator(); it1.hasNext();) {
			for (Iterator<Web2cComponent> it2 = (((Web2cEmbeddedPage)it1.next()).getComponents()).iterator(); it2.hasNext();) {
				list.add(it2.next());
			}
		}
		
		return list;
	}
	
	/**
	 * Returns page's HTML tags
	 * @return HTML tags
	 */
	public ArrayList<String> getHTML() {
		
		ArrayList<String> pageHTML = new ArrayList<String>();	
		
		for(Iterator<ArrayList<String>> it1 = sectionsHTML.iterator(); it1.hasNext();) {
			for(Iterator<String> it2 = ((ArrayList<String>)it1.next()).iterator(); it2.hasNext();) {
				String html = (String)it2.next();
				pageHTML.add(html);
			}
		}
		
		for(Iterator<Web2cEmbeddedPage> it1 = embeddedPages.iterator(); it1.hasNext();) {
			ArrayList<String> embeddedPageHTML = ((Web2cEmbeddedPage)it1.next()).getHTML();
			for(Iterator<String> it2 = embeddedPageHTML.iterator(); it2.hasNext();) {
				String html = (String)it2.next();
				pageHTML.add(html);
			}
		}

		return pageHTML;
	}


	/**
	 * Returns page's JavaScript code snippets
	 * @return JavaScript code snippets
	 */
	public ArrayList<String> getJavaScript() {
		
		ArrayList<String> pageJavaScript = new ArrayList<String>();	
		
		for(Iterator<ArrayList<String>> it1 = sectionsJavaScript.iterator(); it1.hasNext();) {
			for(Iterator<String> it2 = ((ArrayList<String>)it1.next()).iterator(); it2.hasNext();) {
				pageJavaScript.add((String)it2.next());
			}
		}
		
		for(Iterator<Web2cEmbeddedPage> it1 = embeddedPages.iterator(); it1.hasNext();) {
			ArrayList<String> embeddedPageJavaScript = ((Web2cEmbeddedPage)it1.next()).getJavaScript();
			for(Iterator<String> it2 = embeddedPageJavaScript.iterator(); it2.hasNext();) {
				pageJavaScript.add((String)it2.next());
			}
		}

		return pageJavaScript;
	}
	
	/**
	 * Constructs the page
	 */
	private void constructPage() {

		Web2cConfigReader configReader = null;
		
		try {
			configReader = new Web2cConfigReader(new File(homeDir + File.separator + pageConfigFile), debugLogger, debugListener);
			constructSections(configReader.getSectionsAttributeNames(), configReader.getSectionsAttributeValues(), configReader.getSectionsElements(), configReader.getSectionsElementsAttributeNames(), configReader.getSectionsElementsAttributeValues());
			constructEmbeddedPages(configReader.getEmbeddedPagesAttributeNames(), configReader.getEmbeddedPagesAttributeValues());
		}
		catch (Exception e) {
			error = error + configReader.getError();
		}
	}
	
	/**
	 * Constructs the sections of the page
	 * @param sectionsAttributeNames list containing lists of attribute names of all sections on the page
	 * @param sectionsAttributeValues list containing lists of attribute values of all sections on the page
	 * @param sectionsElements list containing lists of element names of all sections on the page
	 * @param sectionsElementsAttributeNames list containing lists containing lists of attribute names of all elements of all sections on the page
	 * @param sectionsElementsAttributeValues list containing lists containing lists of attribute values of all elements of all sections on the page
	 */
	private void constructSections(ArrayList<ArrayList<String>> sectionsAttributeNames, ArrayList<ArrayList<String>> sectionsAttributeValues, ArrayList<ArrayList<String>> sectionsElements, ArrayList<ArrayList<ArrayList<String>>> sectionsElementsAttributeNames, ArrayList<ArrayList<ArrayList<String>>> sectionsElementsAttributeValues) {

		Iterator<ArrayList<String>> it1 = sectionsAttributeNames.iterator();
		Iterator<ArrayList<String>> it2 = sectionsAttributeValues.iterator();
		Iterator<ArrayList<ArrayList<String>>> it4 = sectionsElementsAttributeNames.iterator();
		Iterator<ArrayList<ArrayList<String>>> it5 = sectionsElementsAttributeValues.iterator();
		for (Iterator<ArrayList<String>> it3 = sectionsElements.iterator(); it3.hasNext();) {
			ArrayList<String> attributeNames = (ArrayList<String>)it1.next();
			ArrayList<String> attributeValues = (ArrayList<String>)it2.next();
			ArrayList<String> elementsNames = (ArrayList<String>)it3.next();
			ArrayList<ArrayList<String>> elementsAttributeNames = (ArrayList<ArrayList<String>>)it4.next();
			ArrayList<ArrayList<String>> elementsAttributeValues = (ArrayList<ArrayList<String>>)it5.next();
			Web2cSection section = new Web2cSection(referenceId, homeDir, language, xScale, yScale, attributeNames, attributeValues, elementsNames, elementsAttributeNames, elementsAttributeValues, simulation, userPrivilege, commonDeviceAccess, commonVideoConnector, chatChannel, web2cSession, referencePoint, userName, accessMode, constructionProgress);
			sections.add(section);
			sectionsHTML.add(section.getHTML());
			sectionsJavaScript.add(section.getJavaScript());
			error = error + section.getError();
		}
	}
	
	/**
	 * Constructs the embedded pages of the page
	 * @param embeddedPagesAttributeNames list containing lists of attribute names of all embedded pages on the page
	 * @param embeddedPagesAttributeValues list containing lists of attribute values of all embedded pages on the page
	 */
	private void constructEmbeddedPages(ArrayList<ArrayList<String>> embeddedPagesAttributeNames, ArrayList<ArrayList<String>> embeddedPagesAttributeValues) {

		ArrayList<String> attributeNames = new ArrayList<String>();
		ArrayList<String> attributeValues = new ArrayList<String>();
		Iterator<ArrayList<String>> it2 = embeddedPagesAttributeValues.iterator();
		for (Iterator<ArrayList<String>> it1 = embeddedPagesAttributeNames.iterator(); it1.hasNext();) {
			attributeNames = (ArrayList<String>)it1.next();
			attributeValues = (ArrayList<String>)it2.next();		
			int xLeft = 0;
			int yTop = 0;
			String refFile = null;
			String refId = null;
			String refPoint = null;
			Iterator<String> it3 = attributeValues.iterator();
			for (Iterator<String> it4 = attributeNames.iterator(); it1.hasNext();) {
				String name = (String)it4.next();
				String value = (String)it3.next();
				if (name.equals("uri")) {
					refFile = value.substring(7).replace("/", File.separator);
				}
				if (name.equals("id")) {
					if (value.equals("enter Id") || value.equals("") || value.equals(" ")) {
						error = error + "[Component Decoder] Missing or wrong component ID;";
					}
					refId = referenceId + "_" + value;
				}
				if (name.equals("bounds")) {
					xLeft = Integer.valueOf(value.split(",")[0]);
					yTop = Integer.valueOf(value.split(",")[1]);
					refPoint = ((Integer)((int)((Integer.valueOf(referencePoint.split(",")[0]) + xLeft) * xScale))).toString() + "," + ((Integer)((int)((Integer.valueOf(referencePoint.split(",")[1]) + yTop) * yScale))).toString();
				}
			}
			Web2cEmbeddedPage embeddedPage = new Web2cEmbeddedPage(web2cSession, homeDir, refFile, userName, language, refId, refPoint, xScale, yScale, userPrivilege, simulation, commonDeviceAccess, commonVideoConnector, chatChannel, debugLogger, debugListener, accessMode, constructionProgress);
			embeddedPages.add(embeddedPage); 
			error = error + embeddedPage.getError();
		}
	}
	
	/**
	 * Returns page construction error
	 * @return error string
	 */
	public String getError() {
		
		return error;
	}
}
