package de.desy.web2cToolkit.web2c.web2cDeviceAccess;

import java.util.ArrayList;
import java.util.Calendar;

import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cFileLogger;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cKeyBoardListener;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cDeviceAccess;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cGarbageCollection;


/**
 * Common device access
 * @author R.Bacher, DESY
 */
public class Web2cCommonDeviceAccess implements IWeb2cDeviceAccess, IWeb2cConstants, IWeb2cGarbageCollection {
	
	private Web2cTINEDeviceAccess web2cTINE = null;	
	private Web2cDOOCSDeviceAccess web2cDOOCS = null;	
	private Web2cDOOCSDeviceAccess web2cTANGO = null;	
	private Web2cDOOCSDeviceAccess web2cEPICS = null;	
	private Web2cFileLogger debugLogger= null;
	private Web2cFileLogger actionLogger= null;
	private Web2cKeyBoardListener debugFlagListener = null;
	private ArrayList<String> propertyList = null;
	private ArrayList<Integer> numberDeviceAccessRequesters = new ArrayList<Integer>();
	
	private String error = "";

	/**
	 * Common device access: constructor
	 * @param dLogger debug logger
	 * @param aLogger action logger
	 */
	public Web2cCommonDeviceAccess(Web2cKeyBoardListener listener, Web2cFileLogger dLogger, Web2cFileLogger aLogger) {
		
		web2cTINE = new Web2cTINEDeviceAccess(null);
		web2cDOOCS = new Web2cDOOCSDeviceAccess("doocs");
		web2cTANGO = new Web2cDOOCSDeviceAccess("tango");
		web2cEPICS = new Web2cDOOCSDeviceAccess("epics");
		propertyList = new ArrayList<String>();
		debugFlagListener = listener;
		debugLogger = dLogger;
		actionLogger = aLogger;
	}
	
	/**
	 * Checks whether monitor link for the property specified already exists
	 * @param property property
	 */
	private synchronized boolean monitorIsRegistered(String property) {
		
		boolean isRegistered = false;
		
		synchronized(propertyList) {
			synchronized (numberDeviceAccessRequesters) {	
				if (propertyList.isEmpty()) {
					isRegistered = false;
				}
				else {
					isRegistered = false;
					if (propertyList.contains(property)) isRegistered = true;
				}
			}
		}
		
		return isRegistered;
	}
	
	public String getError() {
		
		String lastError = error;
		error = "";
		
		return lastError;
	}

	public void execute(String property, boolean logMessageEnabled) {
		
		String protocol = "";
		String address = "";
		
		protocol = property.split("://")[0];
		address = property.split("://")[1];
		
		if (logMessageEnabled == true) actionLogger.appendLog(property);
		
		if (protocol.equalsIgnoreCase("tine")) {
			web2cTINE.execute(address, false);
		}		
		if (protocol.equalsIgnoreCase("doocs")) {
			web2cDOOCS.execute(address, false);
		}	
		if (protocol.equalsIgnoreCase("tango")) {
			web2cTANGO.execute(address, false);
		}	
		if (protocol.equalsIgnoreCase("epics")) {
			web2cEPICS.execute(address, false);
		}
	}

	public synchronized String getMonitorData(String property) {
		
		String ret = "";
		String protocol = "";
		String address = "";
		
		protocol = property.split("://")[0];
		address = property.split("://")[1];
		
		if (protocol.equalsIgnoreCase("tine")) {	
			ret = web2cTINE.getMonitorData(address);
		}		
		if (protocol.equalsIgnoreCase("doocs")) {	
			ret = web2cDOOCS.getMonitorData(address);
		}		
		if (protocol.equalsIgnoreCase("tango")) {	
			ret = web2cTANGO.getMonitorData(address);
		}		
		if (protocol.equalsIgnoreCase("epics")) {	
			ret = web2cEPICS.getMonitorData(address);
		}
	    if (debugFlagListener.getDebugControlSystemCommunication()) debugLogger.appendLog("from property " + property + ": " + ret);
	    
		return ret; 
	}

	public synchronized int registerMonitor(String property) {
		
		int status = 0;
		int numberRequesters = 0;
		String protocol = "";
		String address = "";
		
		synchronized(propertyList) {
			synchronized (numberDeviceAccessRequesters) {
				if (!monitorIsRegistered(property.substring(0, property.indexOf("(")))) {	
					protocol = property.split("://")[0];
					address = property.split("://")[1];
			
					if (protocol.equalsIgnoreCase("tine")) {
						status = web2cTINE.registerMonitor(address);		
						error = error + web2cTINE.getError();
					}		
					if (protocol.equalsIgnoreCase("doocs")) {
						status = web2cDOOCS.registerMonitor(address);		
						error = error + web2cDOOCS.getError();
					}		
					if (protocol.equalsIgnoreCase("tango")) {
						status = web2cTANGO.registerMonitor(address);		
						error = error + web2cTANGO.getError();
					}		
					if (protocol.equalsIgnoreCase("epics")) {
						status = web2cEPICS.registerMonitor(address);		
						error = error + web2cEPICS.getError();
					}
			
					if (status == 0) {
						if (!monitorIsRegistered(property.substring(0, property.indexOf("(")))) {
							propertyList.add(property.substring(0, property.indexOf("(")));
							numberDeviceAccessRequesters.add(1);
							System.out.println(String.valueOf(Calendar.getInstance().getTime().toString()) + ": monitor link for property " + property + " sucessfully registered");
						}
						else {
							int index = propertyList.indexOf(property.substring(0, property.indexOf("(")));
							numberRequesters = numberDeviceAccessRequesters.get(index) + 1;
							numberDeviceAccessRequesters.set(index, numberRequesters);
						}
					}
				}
				else {
					int index = propertyList.indexOf(property.substring(0, property.indexOf("(")));
					numberRequesters = numberDeviceAccessRequesters.get(index) + 1;
					numberDeviceAccessRequesters.set(index, numberRequesters);
				}
			}
		}
		
		return status;
	}

	/**
	 * Closes the monitor link specified
	 * @param property monitor id
	 */
	public void closeMonitorLink(String property) {	
		
		String protocol = "";
		String address = "";
		
		protocol = property.split("://")[0];
		address = property.split("://")[1];	
		
		if (protocol.equalsIgnoreCase("tine")) {
			web2cTINE.closeMonitorLink(address);
		}	
		if (protocol.equalsIgnoreCase("doocs")) {
			web2cDOOCS.closeMonitorLink(address);
		}
		if (protocol.equalsIgnoreCase("tango")) {
			web2cTANGO.closeMonitorLink(address);
		}
		if (protocol.equalsIgnoreCase("epics")) {
			web2cEPICS.closeMonitorLink(address);
		}
	}

	public String getHistory(String property) {			
		
		String ret = "";
		String protocol = "";
		String address = "";
		
		protocol = property.split("://")[0];
		address = property.split("://")[1];
	
		if (protocol.equalsIgnoreCase("tine")) {
			ret = web2cTINE.getHistory(address);
		}	
		if (protocol.equalsIgnoreCase("doocs")) {
			ret = web2cDOOCS.getHistory(address);
		}
		if (protocol.equalsIgnoreCase("tango")) {
			ret = web2cTANGO.getHistory(address);
		}
		if (protocol.equalsIgnoreCase("epics")) {
			ret = web2cEPICS.getHistory(address);
		}
		
	    if (debugFlagListener.getDebugControlSystemCommunication()) debugLogger.appendLog("from property " + property + ": " + ret);
	    
		return ret;
	}

	public String getActivityCounter() {
		
		String activityCounter = "Last Device Update:";
		
		if (!web2cTINE.getActivityCounter().equals("")) {
			activityCounter = activityCounter + web2cTINE.getActivityCounter();
		}		
		if (!web2cDOOCS.getActivityCounter().equals("")) {
			activityCounter = activityCounter + web2cDOOCS.getActivityCounter();
		}	
		if (!web2cTANGO.getActivityCounter().equals("")) {
			activityCounter = activityCounter + web2cTANGO.getActivityCounter();
		}	
		if (!web2cEPICS.getActivityCounter().equals("")) {
			activityCounter = activityCounter + web2cEPICS.getActivityCounter();
		}
		
		return activityCounter;
	}
	
	public String getPropertyDescription(String property) {		
		
		String ret = "";	
		String protocol = "";
		String address = "";
		
		protocol = property.split("://")[0];
		address = property.split("://")[1];
	
		if (protocol.equalsIgnoreCase("tine")) {
			ret = web2cTINE.getPropertyDescription(address);
		}	
		if (protocol.equalsIgnoreCase("doocs")) {
			ret = web2cDOOCS.getPropertyDescription(address);
		}
		if (protocol.equalsIgnoreCase("tango")) {
			ret = web2cTANGO.getPropertyDescription(address);
		}
		if (protocol.equalsIgnoreCase("epics")) {
			ret = web2cEPICS.getPropertyDescription(address);
		}
		
		return ret;
		
	}
	
	/**
	* Returns the property list
	* @return number of requesters
	*/
	public ArrayList<String> getPropertyList() {
			
		return propertyList;
	}
	
	/**
	* Returns the number-of-requester list
	* @return number of requesters
	*/
	public ArrayList<Integer> getNumberOfRequesterList() {
			
		return numberDeviceAccessRequesters ;
	}
	
	/**
    * Returns the number of requesters requesting the device access specified
    * @param id device access id
    * @return number of requesters
    */
	public int getNumberOfRequesters(String id) {
		
		int index = 0;
		int number = 0;

		synchronized(propertyList) {
			synchronized (numberDeviceAccessRequesters) {
				index = propertyList.indexOf(id.substring(0, id.indexOf("(")));
		
				if (index >= 0) {
					number = numberDeviceAccessRequesters.get(propertyList.indexOf(id.substring(0, id.indexOf("("))));
				}
				else {
					number = -1;
				}
			}
		}
		
		return number;
	}
	
	/**
    * Decreases (-1) the number of requesters requesting the device access specified; removes the corresponding device access if number <= 0
    * @param id device access id
    */
	public void removeRequester(String id) {
		
		int index = 0;
		int number = 0;

		synchronized(propertyList) {
			synchronized (numberDeviceAccessRequesters) {
				index = propertyList.indexOf(id.substring(0, id.indexOf("("))); 		
				if (index >= 0) {
					number = getNumberOfRequesters(id);
					numberDeviceAccessRequesters.set(propertyList.indexOf(id.substring(0, id.indexOf("("))), number - 1);
//					System.out.println("Device access requester for " + (id.substring(0, id.indexOf("("))) + " gone! Remaining " + getNumberOfRequesters(id) + " requester(s).");
		
					if (number - 1 <= 0) removeResource(id);
				}
			}
		}
	}
	
	/**
    * Removes the device access with the id specified and all associated resources
    * @param id device access id
    */
	public void removeResource(String id) {
		
		int index = 0;

		synchronized(propertyList) {
			synchronized (numberDeviceAccessRequesters) {
				index = propertyList.indexOf(id.substring(0, id.indexOf("(")));
				if (index >= 0) {	
					closeMonitorLink(id.substring(0, id.indexOf("(")));
					propertyList.remove(index);
					numberDeviceAccessRequesters.remove(index);
//					System.out.println("Device access for " + (id.substring(0, id.indexOf("("))) + " closed!");
				}
			}
		}
	}
}
