package de.desy.web2cToolkit.web2c.web2cDeviceAccess;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;

import de.desy.web2cToolkit.web2cUtil.web2cEscape.Web2cASCIICharacterEscaper;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cDeviceAccess;
import ttf.doocs.clnt.EqAdr;
import ttf.doocs.clnt.EqCall;
import ttf.doocs.clnt.EqCallback;
import ttf.doocs.clnt.EqData;

/**
 * Web2c DOOCS (DOOCS->TANGO, DOOCS->EPICS) device access plug
 * @author R.Bacher, DESY
 */
public class Web2cDOOCSDeviceAccess implements IWeb2cDeviceAccess, IWeb2cConstants {
	
	private ArrayList<String> monitorPropertyList = null;
	private ArrayList<EqCall> callList = null;
	private ArrayList<EqAdr> addressList = null;
	private ArrayList<Monitor> monitorList = null;
	private ArrayList<long[][]> historyTimeStampList = null;
	private ArrayList<Object[][]> historyDataList = null;
	
	private String error = null;
	private String activityStatus = "";
	private String protocolId = "";
	private int pollInterval = 1000;
	private Web2cASCIICharacterEscaper characterEscaper = null;

	/**
	 * Web2c DOOCS device access plug: constructor
	 * @param protocol protocol id 
	 */
	public Web2cDOOCSDeviceAccess(String protocol) {
		
		protocolId = protocol;
		monitorPropertyList = new ArrayList<String>();
		callList = new ArrayList<EqCall>();
		addressList = new ArrayList<EqAdr>();
		monitorList = new ArrayList<Monitor>();
		historyTimeStampList = new ArrayList<long[][]>();
		historyDataList = new ArrayList<Object[][]>();

		characterEscaper = new Web2cASCIICharacterEscaper();
	} 
	
	/**
	 * Returns error string
	 * @return error
	 */
	public synchronized String getError() {
		
		String lastError = error;
		error = "";
		
		return lastError;
	} 
	
	/**
	 * Closes the monitor link specified
	 * @param address monitor id
	 */
	public synchronized void closeMonitorLink(String address) {
		
		int index = 0;
		EqCall call = null;
		EqAdr adr = null;
		
		index = monitorPropertyList.indexOf(address);
		call = callList.get(index);
		adr = addressList.get(index);
		int ret = call.clear_monitor(adr);
		if (ret == 0) {		
			if (protocolId.equals("tango")) {
//				System.out.println("DOOCS->TANGO Monitor Link canceled: " + address);
			}			
			else if (protocolId.equals("epics")) {
//				System.out.println("DOOCS->EPICS Monitor Link canceled: " + address);
			}
			else {
//				System.out.println("DOOCS Monitor Link canceled: " + address);
			}
		}
		else {
			if (protocolId.equals("tango")) {
				System.out.println("Could not cancel DOOCS->TANGO monitor mink: " + address);
			}			
			else if (protocolId.equals("epics")) {
				System.out.println("Could not cancel DOOCS->EPICS monitor mink: " + address);
			}
			else {
				System.out.println("Could not cancel DOOCS monitor mink: " + address);
			}
		}
		monitorPropertyList.remove(index);
		callList.remove(index);
		addressList.remove(index);
		monitorList.remove(index);
		historyTimeStampList.remove(index);
		historyDataList.remove(index);
	}

	/**
	 * Executes a synchronous write call
	 * @param address address
	 * @param messageEnabled action message enabled flag
	 */
	public synchronized void execute(String address, boolean messageEnabled) {

		EqCall dCall = null;
		EqAdr dAddress = null;
		EqData dDataOut = null;
		EqData dDataIn = null;
		String parameter = null;
		String property = null;
		String format = "";
		
		property = getLinkAddress(address);
		parameter = getLinkParameters(address).split(";")[0];	
		
		if (protocolId.equals("tango")) {
			dAddress = new EqAdr("tango://" + property);
		}		
		else if (protocolId.equals("epics")) {	
			dAddress = new EqAdr("epics://" + property);
		}
		else {
			dAddress = new EqAdr(property);
		}	
	    dCall = new EqCall();
	    dDataOut = new EqData();
	    dDataIn = new EqData();
	    
	    dDataOut = dCall.set(dAddress, dDataIn);
	    format = dDataOut.type_string();
	    
		if (dDataOut.error() != 0) {		
			if (protocolId.equals("tango")) {
				System.out.println("Could not execute synchronous DOOCS->TANGO call: " + dAddress + ", Data Type: " + format);
			}			
			else if (protocolId.equals("epics")) {
				System.out.println("Could not execute synchronous DOOCS->EPICS call: " + dAddress + ", Data Type: " + format);
			}
			else {
				System.out.println("Could not execute synchronous DOOCS call: " + dAddress + ", Data Type: " + format);
			}
		}
        else {
			parameter = characterEscaper.decodeString(parameter);
		    if (format.equals("INT")) {		
	        	dDataIn.set(Integer.valueOf(parameter));
    		}
    		else if(format.equals("FLOAT")) {	
	        	dDataIn.set(Float.valueOf(parameter));
    		}
    		else if(format.equals("DOUBLE")) {	
	        	dDataIn.set(Double.valueOf(parameter));
    		}
    		else if(format.equals("STRING")) {		
	        	dDataIn.set(parameter);		
    		}
    		else if(format.equals("BOOL")) {	
    			if (parameter.equals("true")) {
    				dDataIn.set_bool(1);		
    			}
    			else {
    				dDataIn.set_bool(0);
    			}
    		}
            dDataOut = dCall.get(dAddress, dDataIn);		
            if (dDataOut.error() != 0) {
    			if (protocolId.equals("tango")) {
    				System.out.println("Could not execute synchronous DOOCS->TANGO call: " + dAddress + ", Data Type: " + format + ", Error: " + dDataOut.error());
    			}			
    			else if (protocolId.equals("epics")) {
    				System.out.println("Could not execute synchronous DOOCS->EPICS call: " + dAddress + ", Data Type: " + format + ", Error: " + dDataOut.error());
    			}
    			else {
    				System.out.println("Could not execute synchronous DOOCS call: " + dAddress + ", Data Type: " + format + ", Error: " + dDataOut.error());
    			}
    		}
        }
	}
	
	/**
	 * Returns local history data
	 * @param address address
	 * @return data
	 */
	public synchronized String getHistory(String address) {
		
		String out = "";
		StringBuffer outBuf = null;
		String property = "";
		int[] index = null;
		
		index = new int[2];

		property = getLinkAddress(address);
		if (getLinkParameters(address).equals("")) {
			index[0] = 0;
			index[1] = index[0];
		}
		else {
			if ((getLinkParameters(address).split(";")).length == 1) {
				index[0] = Integer.valueOf(getLinkParameters(address).split(";")[0]);
				index[1] = index[0];
			}
			else {
				index[0] = Integer.valueOf(getLinkParameters(address).split(";")[0]);
				index[1] = Integer.valueOf(getLinkParameters(address).split(";")[1]);
			}
		}
		
		outBuf = new StringBuffer("");
		synchronized (monitorList) {
			synchronized (historyTimeStampList) {
				synchronized (historyDataList) {
					Iterator<Monitor> it4 = monitorList.iterator();
					Iterator<Object[][]> it3 = historyDataList.iterator();
					Iterator<long[][]> it2 = historyTimeStampList.iterator();
					for (Iterator<String> it1 = monitorPropertyList.iterator(); it1.hasNext();) {
						String prop = (String)(it1.next());
						long[][] hTimeStamp = (long[][])(it2.next());
						Object[][] hData= (Object[][])(it3.next());
						int pointer = ((Monitor)(it4.next())).getArchiveDataPointer();
						if (prop.equals(property)) {
							for (int i = pointer + 1; i < Integer.valueOf(historyInterval); i++) {
								if (hData[index[0]][i] == null)  {hData[index[0]][i] = characterEscaper.encodeString("0.0");}
								outBuf.append("<value>" + hData[index[0]][i] + "</value><timeStamp>" + hTimeStamp[index[0]][i] + "</timeStamp>");
							}
							for (int i = 0; i <= pointer; i++) {
								if (hData[index[0]][i] == null)  {hData[index[0]][i] = characterEscaper.encodeString("0.0");}
								outBuf.append("<value>" + hData[index[0]][i] + "</value><timeStamp>" + hTimeStamp[index[0]][i] + "</timeStamp>");
							}
						}
					}
				}
			}
		}
		out = outBuf.toString();
		
		return out;
	}

	/**
	 * Returns monitor data
	 * @param address address
	 * @return data
	 */
	public synchronized String getMonitorData(String address) {

		String out = null;
		String data = null;
		StringBuffer dataBuf = null;
		String property = "";
		String timeStamp = null;
		int[] index = null;
		int i = 0;
		
		index = new int[2];

		property = getLinkAddress(address);
		if (getLinkParameters(address).equals("")) {
			index[0] = 0;
			index[1] = index[0];
		}
		else {
			if ((getLinkParameters(address).split(";")).length == 1) {
				index[0] = Integer.valueOf(getLinkParameters(address).split(";")[0]);
				index[1] = index[0];
			}
			else {
				index[0] = Integer.valueOf(getLinkParameters(address).split(";")[0]);
				index[1] = Integer.valueOf(getLinkParameters(address).split(";")[1]);
			}
		}
		
		dataBuf = new StringBuffer("");
		Iterator<Monitor> it2 = monitorList.iterator();
		for (Iterator<String> it1 = monitorPropertyList.iterator(); it1.hasNext();) {
			String prop = (String)(it1.next());
			Monitor monitor = (Monitor)(it2.next());
			if (prop.equals(property)) {  
				dataBuf.append("<value>");
				try {	  
					String[] retData = ((monitor.getData()).split(";")[0]).split("<");
					for (i = index[0]; i <= index[1]; i++) {
						dataBuf.append(retData[i] + ",");
					}
					timeStamp = (monitor.getData()).split(";")[1];
				}
				catch (Exception e) {
//					System.out.println("Error getting monitor data: " + prop);
				}
				data = dataBuf.toString();
			    out = data.substring(0, data.length() - 1) + "</value><timeStamp>" + timeStamp + "</timeStamp>";
			}
		}

		return out;
	}

	/**
	 * Registers monitor link
	 * @param address address
	 * @return status
	 */
	public synchronized int registerMonitor(String address) {

		Monitor monitor = null;	
	    EqCall dCall = null;
	    EqData dDataIn = null;
	    EqData dDataOut = null;
	    EqAdr dAddress = null;
		String format = "";
		int size = 0;
		String property = null;
		long[][] hTimeStamp = null;
		Object[][] hData = null;
		int status = 0;
		
		property = getLinkAddress(address);
		
		status = 0;
		try {		    
			dCall = new EqCall();
			dDataIn = new EqData();
			dDataOut = new EqData();
			if (protocolId.equals("tango")) {
				dAddress = new EqAdr("tango://" + property);
			}		
			else if (protocolId.equals("epics")) {	
				dAddress = new EqAdr("epics://" + property);
			}
			else {
				dAddress = new EqAdr(property);
			}

		    dDataOut = dCall.get(dAddress, dDataIn); 
		    format = dDataOut.type_string();
		    size = dDataOut.array_length(); 
		    if (size == 0) size = 1;
			if (size > maxLinkDataSize) size = maxLinkDataSize;

			if (dDataOut.error() == 0) { 
				System.out.println(property);
				monitorPropertyList.add(property);
				hTimeStamp = new long[size][Integer.valueOf(historyInterval)];
				historyTimeStampList.add(hTimeStamp);
				hData = new Object[size][Integer.valueOf(historyInterval)];
				historyDataList.add(hData);
				monitor = new Monitor(address, format, size, hTimeStamp, hData);
				monitorList.add(monitor);	
				callList.add(dCall);
				addressList.add(dAddress);
				dCall.get_monitor(dAddress, dDataIn, dDataOut, EqCall.MODE_RATE, 1000.0f / pollInterval, monitor, address);
    			if (protocolId.equals("tango")) {
    				System.out.println(String.valueOf(Calendar.getInstance().getTime().toString()) + " DOOCS->TANGO monitor link opened: " + dAddress.show_adr() + ", Data Type: " + format + ", Data Size: " + size);
    			}			
    			else if (protocolId.equals("epics")) {
    				System.out.println(String.valueOf(Calendar.getInstance().getTime().toString()) + " DOOCS->EPICS monitor link opened: " + dAddress.show_adr() + ", Data Type: " + format + ", Data Size: " + size);
    			}
    			else {
    				System.out.println(String.valueOf(Calendar.getInstance().getTime().toString()) + " DOOC monitor link opened: " + dAddress.show_adr() + ", Data Type: " + format + ", Data Size: " + size);
    			}
			}
			else {
				status = status + 1;
    			if (protocolId.equals("tango")) {
    				System.out.println("Could not establish DOOCS->TANGO monitor link: " + dAddress.show_adr() + ", Data Type: " + format + ", Error: " + dDataOut.error());
    			}			
    			else if (protocolId.equals("epics")) {
    				System.out.println("Could not establish DOOCS->EPICS monitor link: " + dAddress.show_adr() + ", Data Type: " + format + ", Error: " + dDataOut.error());
    			}
    			else {
    				System.out.println("Could not establish DOOCS monitor link: " + dAddress.show_adr() + ", Data Type: " + format + ", Error: " + dDataOut.error());
    			}
				error = error + "[Control system] Error establishing monitor link " + dAddress.show_adr() + ";";
			}
		}
		catch (Exception e) {
			status = status + 1;
			error = error + "[Control system] Error establishing monitor link " + dAddress.show_adr() + ";";
		}
		
		return status;
	}
	
	/**
	 * DOOCS Monitor
	 */
	public class Monitor implements EqCallback {
		
		String monitorProperty = null;
		String linkFormat = "";
		int linkSize = 0;
		long[][] hTimeStamp = null;
		Object[][] hData = null;
		String linkData = "";
		long timeStamp = 0;			
		int[] index = null;
		String errorData = "";
		StringBuffer errorDataBuf = new StringBuffer("");
		int archiveDataPointer = 0;
		
		/**
		 * Constructor
		 * @param address link address
		 * @param format link format
		 * @param size link size
		 * @param timeStamp time stamp array for local history
		 * @param data data array for local history
		 */
		public Monitor (String address, String format, int size, long[][] timeStamp, Object[][] data) {

			monitorProperty = address;
			linkFormat = format;
			linkSize = size;
			hTimeStamp = timeStamp;
			hData = data;
			if(linkFormat.equals("STRING") || linkFormat.equals("STRING16") || linkFormat.equals("A_USTR")) {	
				for (int i = 0; i < linkSize; i++) {	
					errorDataBuf.append(characterEscaper.encodeString("?") + "<");
				}
			}
			else {
				for (int i = 0; i < linkSize; i++) {	
					errorDataBuf.append(characterEscaper.encodeString("0") + "<");
				}
			}	
			errorData = errorDataBuf.toString();
			errorData = errorData.substring(0, errorData.length() - 1);
			linkData = errorData + ";" + String.valueOf(Calendar.getInstance().getTimeInMillis() / 1000); 	
			
			long startTime = System.currentTimeMillis();
			for (int i = 0; i < linkSize; i++) {
				for (int j = 0; j < Integer.valueOf(historyInterval); j++) {
					hTimeStamp[i][j] = startTime - ((Integer.valueOf(historyInterval) - j) * pollInterval);
					hData[i][j] = characterEscaper.encodeString("0.0");
				}
			}
		}
		
		/**
		 * Executed if called back by DOOCS lib
		 * @param data out-data object 
		 * @param object user-data object (== link address), used as link id
		 */
		public synchronized void dataCallback(EqData data, Object object) {
			
			int linkStatus = 0;
			String outData = "";
			StringBuffer outDataBuf = new StringBuffer("");
			
			if (monitorProperty.equals((String)object)) {
				
				linkStatus = data.error();
				linkFormat = data.type_string();
				timeStamp = Calendar.getInstance().getTimeInMillis();
				
				synchronized (activityStatus) {
					if (protocolId.equals("tango")) {
						activityStatus = " [DOOCS->TANGO] " + "Error: " + linkStatus + " (" + timeStamp + ")";
					}
					else if (protocolId.equals("epics")) {
						activityStatus = " [DOOCS0->EPICS] " + "Error: " + linkStatus + " (" + timeStamp + ")";
					}
					else {
						activityStatus = " [DOOCS] " + "Error: " + linkStatus + " (" + timeStamp + ")";
					}
				}
				
				if (linkStatus != 0) {
//					System.out.println("Link Error : " + monitorProperty + " " + "Error: " + linkStatus);
				}
				else {
					if (data.toString().equals("") || data.equals(null)) {
						if(linkFormat.equals("STRING") || linkFormat.equals("STRING16") || linkFormat.equals("A_USTR")) {	
							for (int i = 0; i < linkSize; i++) {	
								outDataBuf.append(characterEscaper.encodeString("?") + "<");
							}
						}
						else {
							for (int i = 0; i < linkSize; i++) {	
								outDataBuf.append(characterEscaper.encodeString("0") + "<");
							}
						}	
					}
					else {
						if (linkFormat.equals("INT") || linkFormat.equals("A_INT") || linkFormat.equals("BOOL")) {					
							for (int i = 0; i < linkSize; i++) {
								outDataBuf.append(characterEscaper.encodeString(String.valueOf(data.get_int(i))) + "<");
							}		
						}
						else if(linkFormat.equals("FLOAT") || linkFormat.equals("A_FLOAT") || linkFormat.equals("DOUBLE") ||
								linkFormat.equals("TDS") || linkFormat.equals("A_TDS") || linkFormat.equals("SPECTRUM")) {					
							for (int i = 0; i < linkSize; i++) {
								outDataBuf.append(characterEscaper.encodeString(String.valueOf(data.get_float(i))) + "<");
							}		
						}
						else if(linkFormat.equals("STRING") || linkFormat.equals("STRING16")) {						
							for (int i = 0; i < linkSize; i++) {
								outDataBuf.append(characterEscaper.encodeString(String.valueOf(data.get_string(i)).trim()) + "<");
							}			
						}
						else if(linkFormat.equals("A_USTR")) {						
							for (int i = 0; i < linkSize; i++) {
								String string = new String(data.get_ustr(i).str_data);
								outDataBuf.append(characterEscaper.encodeString(string.trim()) + "<");
							}	
						}
						else if(linkFormat.equals("A_BYTE")) {						
							for (int i = 0; i < linkSize; i++) {
								outDataBuf.append(characterEscaper.encodeString(String.valueOf(data.get_byte(i))) + "<");
							}	
						}	
					}
					outData = outDataBuf.toString();
					outData = outData.substring(0, outData.length() - 1);
					linkData = outData + ";" + String.valueOf(timeStamp);

					if (linkFormat.equals("INT") || linkFormat.equals("A_INT") || linkFormat.equals("BOOL") || 
						linkFormat.equals("FLOAT") || linkFormat.equals("A_FLOAT") || linkFormat.equals("DOUBLE") || 
						linkFormat.equals("TDS") || linkFormat.equals("A_TDS") || linkFormat.equals("SPECTRUM") || 
						linkFormat.equals("BYTE")) {
						String[] ldata = linkData.split(";")[0].split("<");
						for (int j = 0; j < linkSize; j++) {
							hTimeStamp[j][archiveDataPointer] = timeStamp;
							hData[j][archiveDataPointer] = (Object)(ldata[j]);
						}
						archiveDataPointer = archiveDataPointer + 1;
						if (archiveDataPointer == Integer.valueOf(historyInterval)) {
							archiveDataPointer = 0;
						}
					}
				}
			}
	    }
		
		/**
		 * Returns actual link data
		 * @return data 
		 */
		public synchronized String getData() {
			
			return linkData; 
		}
		
		/**
		 * Returns actual archive data pointer
		 * @return pointer 
		 */
		public synchronized int getArchiveDataPointer() {
			
			return archiveDataPointer; 
		}
	}
	
	private synchronized String getLinkAddress(String address) {
		
		String property = null;
		
		property = address.substring(0,address.indexOf("("));
		
		return property;
	}
	
	private synchronized String getLinkParameters(String address) {

		String parameters = "";

		if (address.substring(address.indexOf("("),address.indexOf(")") + 1).length() != 2) {
			parameters = address.substring(address.indexOf("(") + 1, address.indexOf(")"));
		}
		else {
			parameters = "";
		}

		return parameters;
	}

	/**
	 * Returns activity status
	 * @return activity status
	 */
	public synchronized String getActivityCounter() {
		return activityStatus;   // "[DOOCS, DOOCS->TANGO, DOOCS->EPICS] errorCode (timeStamp)"
	}
	
	/**
	 * Returns property description
	 * @param address address
	 * @return description
	 */
	public String getPropertyDescription(String address) {	
		
		return "";
	}
}
