package de.desy.web2cToolkit.web2cMessenger;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cFileLogger;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;

/**
 * @author R.Bacher, DESY
 */
public class Web2cSMTPInterface implements IWeb2cConstants {

	private Session smtpSession = null;
	private MimeMessage smtpMessage = null;
	private Transport transport = null;
	private Properties properties = null;	
	private String id = "";
	private String password = "";
	private String host = "";
	private String logging = "no";
	private boolean useCredentials = true;
	private boolean serviceAvailable = false;
	private boolean loggingEnabled = false;
	
	private String initFileName = "";
	private Web2cFileLogger smtpLogger = null; 
	private boolean iniFileRead = false;
	private String status = "Status: Undefined";
	
    /**
     * SMTP interface (e.g. E-Mail or SMS)
     * Provides on-request messaging 
     * Constructor
     * @param name init file name
	 * @param logger logger
     */
	public Web2cSMTPInterface(String name, Web2cFileLogger logger) {

		initFileName = name;
		smtpLogger = logger;
		readInitializationFile();
	}	
	
	/**
	* Arms the service
	*/
	public void armService() {

		initializeService();
	}
	
	/**
	* Disarms the service
	*/
	public void disarmService() {

		try {
			transport.close();
			smtpSession = null;
			smtpMessage = null;
			transport = null;
			iniFileRead = false;
			serviceAvailable = false;	
		} catch (MessagingException e) {
			System.out.println("Error disarming SMTP service.");
		}
	}
	
	/**
	* Returns the service status.
	* @return status
	*/
	public boolean getServiceAvailable() {
			
		return serviceAvailable;
	}	
	
	/**
	* Returns the last smtp message sending status.
	* @return status
	*/
	public String getMessageSendingStatus() {
			
		return status;
	}	
	
	/**
	* Returns the service id.
	* @return id
	*/
	public String getServiceId() {
			
		return id;
	}			
	
	/**
	* Sets the service id.
	* @param text id
	*/
	public void setServiceId(String text) {
		
		id = text;
	}				
	
	/**
	* Saves the service id.
	* @param text id
	*/
	public void saveServiceId(String text) {
		
		id = text;
		writeInitializationFile();
	}
	
	/**
	* Returns the service password.
	* @return password
	*/
	public String getServicePassword() {
			
		return password;
	}				
	
	/**
	* Sets the service password.
	* @param text password
	*/
	public void setServicePassword(String text) {
		
		password = text;
	}					
	
	/**
	* Saves the service password.
	* @param text password
	*/
	public void saveServicePassword(String text) {
		
		password = text;
		writeInitializationFile();
	}
	
	/**
	* Returns the service host name.
	* @return host name
	*/
	public String getServiceHostName() {
			
		return host;
	}				
	
	/**
	* Sets the service host name.
	* @param text host name
	*/
	public void setServiceHostName(String text) {
		
		host= text;
	}					
	
	/**
	* Saves the service host name.
	* @param text host name
	*/
	public void saveServiceHostName(String text) {
		
		host = text;
		writeInitializationFile();
	}
	
	/**
	* Returns the service logging-option.
	* @return option
	*/
	public String getServiceLoggingOption() {
			
		return logging;
	}		
	
	/**
	* Sets the service logging-option.
	* @param text option
	*/
	public void setServiceLoggingOption(String text) {
			
		logging = text;
	}		
	
	/**
	* Saves the service logging-option.
	* @param text option
	*/
	public void saveServiceLoggingOption(String text) {
			
		logging = text;
		writeInitializationFile();
	}
	
    /**
     * Write the Web2cMessenger initialization file
     */
	private void writeInitializationFile() {
		
	    try {
	    	BufferedWriter out = new BufferedWriter(new FileWriter(registryDir + File.separator + initFileName));
	    	out.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"); out.newLine();
	    	out.write("<messenger>"); out.newLine();
	    	out.write("   <interface>smtp</interface>"); out.newLine();
	    	out.write("   <id>" + id + "</id>"); out.newLine();
	    	out.write("   <password>" + password + "</password>"); out.newLine();
	    	out.write("   <host>" + host + "</host>"); out.newLine();
	    	out.write("   <logging>" + logging + "</logging>"); out.newLine();
	    	out.write("</messenger>"); out.newLine();
			out.flush();
			out.close();
		} catch (IOException e) {
			System.out.println("Error writing Web2cSMTPInit file.");
		}
	}
	
    /**
     * Reads the Web2cMessenger initialization file
     */
	@SuppressWarnings("unchecked")
	private void readInitializationFile() {
		
	   	SAXBuilder builder = null;
    	Document profile = null;
    	Element messenger = null;
    	List<Element> messengerAttributeList = null;
    	Element messengerAttribute = null;

		try {
			id = "";
			password = "";
			host = "";
			logging = "";
			builder = new SAXBuilder();
			profile = builder.build(new File(registryDir + File.separator + initFileName));
			messenger = profile.getRootElement();
			messengerAttributeList = (List<Element>)messenger.getChildren();    
			for (Iterator<Element> it1 = messengerAttributeList.iterator(); it1.hasNext();) {
				messengerAttribute = (Element)it1.next(); 	
				if ((messengerAttribute.getName()).equals("id")) {
					id = messengerAttribute.getValue();
				} 
				else if ((messengerAttribute.getName()).equals("password")) {
					password = messengerAttribute.getValue();
				}
				else if ((messengerAttribute.getName()).equals("host")) {
					host = messengerAttribute.getValue();
				}
				else if ((messengerAttribute.getName()).equals("logging")) {
					logging = messengerAttribute.getValue();
				}
			}

			properties = System.getProperties();
			properties.put("mail.host", host);
			properties.put("mail.transport.protocol", "smtp");	
			if (!id.equals("") && !id.equals(" ")) {
				properties.put("mail.smtp.auth", "true");
				useCredentials = true;
			}
			else {
				properties.put("mail.smtp.auth", "false");
				useCredentials = false;
			}

			if (logging.equals("yes")) {
				loggingEnabled = true;
			}
			else {
				loggingEnabled = false;
			}
			iniFileRead = true;
		} 
		catch (JDOMException e) {
			iniFileRead = false;
			System.out.println("Error parsing SMTP profile file! SMTP service not available!");
		} 
		catch (IOException e) {
			iniFileRead = false;
			System.out.println("Error reading SMTP profile file! SMTP service not available!");
		}
	}    
	
	/**
     * Initializes the SMTP service
     */
	private void initializeService() {

		if (iniFileRead == true) {		
			smtpSession = Session.getDefaultInstance(properties, null);
			smtpMessage = new MimeMessage(smtpSession);	
			try {
				if (useCredentials == true) {
					transport = smtpSession.getTransport("smtp");
					transport.connect(host, id, password);
					smtpMessage.saveChanges();
					serviceAvailable = true;
				}
				else {
					serviceAvailable = true;
				}
			}
			catch (Exception e) {
				serviceAvailable = false;
				System.err.println(e.toString());
				e.printStackTrace();
				System.out.println("SMTP service not available!");
			}
		}
		else {
			serviceAvailable = false;
			System.out.println("SMTP service not available!");
		}
	}
	
	/**
	 * Sends a SMTP message
	 * @param from sender's address
	 * @param to recipient's a
	 * @param subject subject text
	 * @param message message text
	 * @return send status message
	 */
	public String sendMessage (String from, String to, String subject, String message) {

		InternetAddress[] recipientAddress = new InternetAddress[1];
		Date date = new Date();

		if (serviceAvailable == true) {		   
			try {
				smtpMessage.setFrom(new InternetAddress(from));
				recipientAddress[0] = new InternetAddress(to);
				smtpMessage.setRecipients(Message.RecipientType.TO, recipientAddress);
				smtpMessage.setSubject(subject);
				smtpMessage.setSentDate(date);
				smtpMessage.setText(message);
				status = "Status: Undefined";
				if (useCredentials == true) {
					try {
						transport.sendMessage(smtpMessage, smtpMessage.getAllRecipients());
						status = "Status: SMTP Message sent";
					}
					catch (Exception e) {
						status = "Status: SMTP Message not sent";
						if (loggingEnabled == true) smtpLogger.appendLog("Error sending SMTP message! date: " + date.toString() + " from: " + from + " to: " + to + " subject: " + subject + " message: " + message);
					}
				}
				else {
					Transport.send(smtpMessage, smtpMessage.getAllRecipients());
					status = "Status: SMTP Message sent";
				}
			}
			catch (Exception e) {
				status = "Status: SMTP Message not sent";
				if (loggingEnabled == true) smtpLogger.appendLog("Error sending SMTP message! from: " + from + " to: " + to + " subject: " + subject + " message: " + message);
			}
		}
		else {
			status = "Status: SMTP Message not sent";
			if (loggingEnabled == true) smtpLogger.appendLog("Error sending SMTP message! SMTP service not available!");
		}	
		
		return status;
	}
}