package de.desy.web2cToolkit.web2cMessenger;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Observable;
import java.util.Observer;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import de.desy.web2cToolkit.web2c.web2cDeviceAccess.Web2cCommonDeviceAccess;
import de.desy.web2cToolkit.web2cUtil.web2cDebug.Web2cFileLogger;
import de.desy.web2cToolkit.web2cUtil.web2cEscape.Web2cASCIICharacterEscaper;
import de.desy.web2cToolkit.web2cUtil.web2cFileIO.Web2cMessengerConfigReader;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;
import twitter4j.AsyncTwitter;
import twitter4j.AsyncTwitterFactory;
import twitter4j.ResponseList;
import twitter4j.Status;
import twitter4j.TwitterAdapter;
import twitter4j.TwitterException;
import twitter4j.TwitterMethod;
import twitter4j.http.RequestToken;

/**
 * @author R.Bacher, DESY
 */
public class Web2cTwitterInterface implements IWeb2cConstants, Observer  {

	private TwitterAdapter adapter = null;
    private AsyncTwitterFactory factory = null;
    private AsyncTwitter twitter = null;
    private RequestToken requestToken = null;
    private String tweet = "";
    private long statusId = 0;
    private long oldStatusId = 0;
    private ArrayList<Long> statusIdList = new ArrayList<Long>();
    private int nextTweetIndex = 0;
    
    private String initFileName = "";		
    private String id = "";
	private String password = "";
	private String logging = "no";
    private Web2cCommonDeviceAccess commonDeviceAccess = null;
	private String twitterDir = "";
	private Web2cFileLogger twitterLogger = null;
	private Web2cMessengerConfigReader messengerConfigReader = null;
    
	private ArrayList<String> groupCaption = null;
	private ArrayList<ArrayList<String>> lineText = null;
	private ArrayList<ArrayList<String>> lineUri = null;
	private ArrayList<ArrayList<String>> lineFormat = null;
	ArrayList<String> monitorList = new ArrayList<String>();

	private int resetCounter = 0;
	private int resetCounterMax = 20;
	private boolean serviceAvailable = false;
	private boolean loggingEnabled = false;
	private TwitterUpdater twitterUpdater = null;
	private TwitterConfigFileParser twitterParser = null;
	private Web2cASCIICharacterEscaper characterEscaper = null;
	private boolean iniFileRead = false;
	
    /**
     * Twitter interface
     * Provides a service which publishes periodically predefined twitter statuses based on control system readings
     * Constructor
     * @param name initialization file name
	 * @param cda actual common device access instance
	 * @param dir directory of file messenger.xml
	 * @param logger logger
     */
	public Web2cTwitterInterface(String name, Web2cCommonDeviceAccess cda, String dir, Web2cFileLogger logger) {
		
		initFileName = name;
		commonDeviceAccess = cda;
		twitterDir = dir;
		twitterLogger = logger;
		characterEscaper = new Web2cASCIICharacterEscaper();	
		readInitializationFile();
	}	
	
	/**
	 * Returns the control system monitors attached to this service
	 * @return list of monitors
	 */
	public ArrayList<String> getMonitors() {
		
		return monitorList;
	}	
	
	/**
	* Arms the service
	*/
	public void armService() {

		initializeService();
	}
	
	/**
	* Disarms the service
	*/
	public void disarmService() {
		
		stopPublishingStatus();
		iniFileRead = false;
		serviceAvailable = false;
		twitterParser = null;
		twitterUpdater = null;
		factory = null;
		adapter = null;
	}
	
	/**
    * Starts publishing the status
    */
	public void startPublishingStatus() {

		if (serviceAvailable == true && twitterUpdater != null) twitterUpdater.startUpdater();
	}
	
	/**
    * Stops publishing the status
    */
	public void stopPublishingStatus() {

		if (serviceAvailable == true && twitterUpdater != null) twitterUpdater.stopUpdater();
	}
	
	/**
	* Returns the last tweet.
	* @return tweet
	*/
	public String getMessage() {
			
		return tweet;
	}		
	
	/**
	* Returns the publishing status.
	* @return status
	*/
	public boolean getServiceStarted() {
		
		boolean status = false;
		
		if (serviceAvailable == true && twitterUpdater != null) status = twitterUpdater.getUpdaterStatus();
		else status = false;
			
		return status;
	}
	
	/**
	* Returns the service status.
	* @return status
	*/
	public boolean getServiceAvailable() {
			
		return serviceAvailable;
	}	
	
	/**
	* Returns the service id.
	* @return id
	*/
	public String getServiceId() {
			
		return id;
	}			
	
	/**
	* Sets the service id.
	* @param text id
	*/
	public void setServiceId(String text) {
		
		id = text;
	}				
	
	/**
	* Saves the service id.
	* @param text id
	*/
	public void saveServiceId(String text) {
		
		id = text;
		writeInitializationFile();
	}
	
	/**
	* Returns the service password.
	* @return password
	*/
	public String getServicePassword() {
			
		return password;
	}				
	
	/**
	* Sets the service password.
	* @param text password
	*/
	public void setServicePassword(String text) {
		
		password = text;
	}					
	
	/**
	* Saves the service password.
	* @param text password
	*/
	public void saveServicePassword(String text) {
		
		password = text;
		writeInitializationFile();
	}
	
	/**
	* Returns the service logging-option.
	* @return option
	*/
	public String getServiceLoggingOption() {
			
		return logging;
	}		
	
	/**
	* Sets the service logging-option.
	* @param text option
	*/
	public void setServiceLoggingOption(String text) {
			
		logging = text;
	}		
	
	/**
	* Saves the service logging-option.
	* @param text option
	*/
	public void saveServiceLoggingOption(String text) {
			
		logging = text;
		writeInitializationFile();
	}
	
	/**
	* Returns the authorization URL.
	*/
	public String getAuthorizationURL() {
			
		return requestToken.getAuthorizationURL();
	}
	
    /**
     * Write the Web2cMessenger initialization file
     */
	private void writeInitializationFile() {
		
	    try {
	    	BufferedWriter out = new BufferedWriter(new FileWriter(registryDir + File.separator + initFileName));
	    	out.write("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"); out.newLine();
	    	out.write("<messenger>"); out.newLine();
	    	out.write("   <interface>twitter</interface>"); out.newLine();
	    	out.write("   <id>" + id + "</id>"); out.newLine();
	    	out.write("   <password>" + password + "</password>"); out.newLine();
	    	out.write("   <logging>" + logging + "</logging>"); out.newLine();
	    	out.write("</messenger>"); out.newLine();
			out.flush();
			out.close();
		} catch (IOException e) {
			System.out.println("Error writing Web2cTwitterInit file.");
		}
	}
	
    /**
     * Reads the Web2cMessenger initialization file
     */
	@SuppressWarnings("unchecked")
	private void readInitializationFile() {
		
	   	SAXBuilder builder = null;
    	Document profile = null;
    	Element messenger = null;
    	List<Element> messengerAttributeList = null;
    	Element messengerAttribute = null;

		try {
			id = "";
			password = "";
			logging = "";
			builder = new SAXBuilder();
			profile = builder.build(new File(registryDir + File.separator + initFileName));
			messenger = profile.getRootElement();
			messengerAttributeList = (List<Element>)messenger.getChildren();    
			for (Iterator<Element> it1 = messengerAttributeList.iterator(); it1.hasNext();) {
				messengerAttribute = (Element)it1.next(); 	
				if ((messengerAttribute.getName()).equals("id")) {
					id = messengerAttribute.getValue();
				} 
				else if ((messengerAttribute.getName()).equals("password")) {
					password = messengerAttribute.getValue();
				}
				else if ((messengerAttribute.getName()).equals("logging")) {
					logging = messengerAttribute.getValue();
				}
			}

			if (logging.equals("yes")) {
				loggingEnabled = true;
			}
			else {
				loggingEnabled = false;
			}
			iniFileRead = true;
		} 
		catch (JDOMException e) {
			iniFileRead = false;
			System.out.println("Error parsing twitter profile file! Twitter service not available!");
		} 
		catch (IOException e) {
			iniFileRead = false;
			System.out.println("Error reading twitter profile file! Twitter service not available!");
		}
	}
	
    /**
     * Initializes the twitter factory
     */
	private void initializeService() {

		if (iniFileRead == true) {		    
			try {
				adapter = new TwitterAdapter() {
					public void gotUserTimeline(ResponseList<Status> statuses) {
						if (statuses.isEmpty() == false) {
							for (int i = 0; i < statuses.toArray().length; i++) {
								Status status = (Status)(statuses.toArray()[i]);
								if (i > 0) {
									if (serviceAvailable == true) twitter.destroyStatus(status.getId());
								}
								if (i == 0 && twitterUpdater.getUpdaterStatus() == false) {
									if (serviceAvailable == true) twitter.destroyStatus(status.getId());
								}
							}
						}
					}
		            public void updatedStatus(Status status) {
		            	if (twitterUpdater.getUpdaterStatus() == true) {
			            	statusId = status.getId();  
			            	if (loggingEnabled == true) twitterLogger.appendLog("Twitter status (" + statusId + ") updated: " + status.getText());
		            		if (!statusIdList.contains(statusId)) {
		            			statusIdList.add(0, statusId);
		            		}
	            			if (statusIdList.size() > groupCaption.size()) {
	            				for (int i = groupCaption.size(); i < statusIdList.size(); i++) {
		            				twitter.destroyStatus(statusIdList.get(i));
	            				}
	            			}
		            	}
		            }
		            public void destroyedStatus(Status status) {
		            	if (twitterUpdater.getUpdaterStatus() == true) {
            				oldStatusId = status.getId();  
            				statusIdList.remove(oldStatusId);
			            	if (loggingEnabled == true) twitterLogger.appendLog("Twitter status (" + oldStatusId + ") destroyed: " + status.getText());
		            	}
		            }
		            public void onException(TwitterException e, TwitterMethod method) {
		            	if (twitterUpdater.getUpdaterStatus() == true) {
			                if (method == TwitterMethod.UPDATE_STATUS) {	
			                	if (loggingEnabled == true) twitterLogger.appendLog("Error updating twitter status: " + e.getMessage());
			                }
			                if (method == TwitterMethod.DESTROY_STATUS) {
			                	if (loggingEnabled == true) twitterLogger.appendLog("Error destroying twitter status: " + e.getMessage());
			                }
		            	}
		                if (method == TwitterMethod.USER_TIMELINE) {
		                	if (loggingEnabled == true) twitterLogger.appendLog("Error destroying user timeline: " + e.getMessage());
		                }
		            }
				};
		        factory = new AsyncTwitterFactory(adapter);
				twitter = factory.getOAuthAuthorizedInstance(id, password);
		    	System.out.println("Get authorization request token ...");
				requestToken = twitter.getOAuthRequestToken();
		    	System.out.println("... token received.");
				System.out.println("Please grant access to your twitter account using URL: " + requestToken.getAuthorizationURL());
				twitterParser = new TwitterConfigFileParser();
				twitterParser.addObserver(this);
				twitterParser.startParser();
		    }
		    catch (TwitterException e) {
				serviceAvailable = false;
				System.err.println(e.toString());
				e.printStackTrace();
				System.out.println("Error getting authorization request token from twitter! Twitter service not available!");
			}
		}
		else {
			serviceAvailable = false;
			System.out.println("Twitter service not available!");
		}
	}
	
	/**
	 * Gets the completion notification from the twitter configuration file parser and starts the twitter update service
	 * @param parser twitter file parser
	 * @param parserStatus error status
	 */
	public void update(Observable parser, Object parserStatus) {

		if (serviceAvailable == true) twitterUpdater = new TwitterUpdater();
	}
	
	/**
	 * @author R.Bacher, DESY
	 */
	public class TwitterConfigFileParser extends Observable implements Runnable {
		
		private Thread parser = null;
		
		/**
		 * Constructor
		 * Parses the twitter configuration file
		 */
		public TwitterConfigFileParser() {
			
			if (parser == null) {
				parser = new Thread(this);
			}
		}
		
		/**
		 * Starts the parser
		 */
		public void startParser() {  
			
			parser.start();
		}
		
		/**
		 * Parses the configuration file containing the content to be sent to twitter and establishes the corresponding monitor links
		 */
		public void run() {  
			
			try {
				groupCaption = new ArrayList<String>();
				lineText = new ArrayList<ArrayList<String>>();
				lineUri = new ArrayList<ArrayList<String>>();
				lineFormat = new ArrayList<ArrayList<String>>();
				
				messengerConfigReader = new Web2cMessengerConfigReader(new File(twitterDir + File.separator + messengerConfigFile));				
				Iterator<String> it3 = messengerConfigReader.getServiceNameList().iterator();
				Iterator<ArrayList<String>> it2 = messengerConfigReader.getServiceConfigValueLists().iterator();
				for (Iterator<ArrayList<String>> it1 = messengerConfigReader.getServiceConfigNameLists().iterator(); it1.hasNext();) {
					ArrayList<String> nameList = (ArrayList<String>)it1.next();
					ArrayList<String> valueList = (ArrayList<String>)it2.next();
					String service = (String)it3.next();
//					System.out.println(service);
					if (service.equals("Twitter")) {
						ArrayList<String> text = new ArrayList<String>();
						ArrayList<String> uri = new ArrayList<String>();
						ArrayList<String> format = new ArrayList<String>();
						Iterator<String> it5 = nameList.iterator();
						for (Iterator<String> it4 = valueList.iterator(); it4.hasNext();) {
							String value = (String)it4.next();
							String name = (String)it5.next();
//							System.out.println(name);
							if (name.equals("caption")) {
								groupCaption.add(value);
//								System.out.println(value);
							} 
							else if (name.equals("line")) {
								int ret = 0;
//								System.out.println(value.split(",")[0]);
//								System.out.println(value.split(",")[1]);
//								System.out.println(value.split(",")[2]);
								ret = commonDeviceAccess.registerMonitor(value.split(",")[2]);
								if (ret == 0) {
									text.add(value.split(",")[0]);
									format.add(value.split(",")[1]);
									uri.add(value.split(",")[2]);
									monitorList.add(value.split(",")[2]);
								}
							}
						}
						lineText.add(text);
						lineFormat.add(format);
						lineUri.add(uri);
					}
				}
				serviceAvailable = true;
			} 
			catch (Exception e) {
				serviceAvailable = false;
				System.err.println(e.toString());
				e.printStackTrace();
				System.out.println("Error parsing twitter file!");
			} 
			
			setChanged();
			notifyObservers(null);
			
			parser.interrupt();   
		}
	}
	
	/**
	 * @author R.Bacher, DESY
	 */
	public class TwitterUpdater implements Runnable {
		
		private Thread updater = null;
		private boolean updaterStarted = false;
		
		/**
		 * Constructor
		 * Update timer
		 */
		public TwitterUpdater() {
			
			if (updater == null) {
				updater = new Thread(this);
				serviceAvailable = true;
				System.out.println("Twitter service available!");
			}
		}
		
		/**
		 * Starts updating the twitter status
		 */
	    public void startUpdater() {
	    	
	    	if (updaterStarted == false) {
		    	try {
		    		twitter.getOAuthAccessToken();
			    	updater.start();
		    	} 
		    	catch (TwitterException e) {
		    		serviceAvailable = false;
					System.out.println("Error getting access token from twitter!");
		    	}
	    	}
	    }
		
		/**
		 * Stops updating the twitter status
		 */
	    public void stopUpdater() {
	    	
	    	updaterStarted = false;
	    }
		
		/**
		 * Returns the updater status
		 * @return updater status
		 */
		public boolean getUpdaterStatus() {
			
			return updaterStarted;
		}

		/**
		 * Updates the twitter status and clears the time line
		 */
		public void run() {    

    		serviceAvailable = true;
	    	updaterStarted = true;
			twitter.getUserTimeline();
			System.out.println("Updating twitter status started!");
			while (updaterStarted == true) {
				try {
					if (updaterStarted == true) {
						for (int i = 0; i < (twitterUpdateInterval / 2); i++) {
							if (updaterStarted == true) Thread.sleep(1000);
						}
					}
					if (updaterStarted == true) generateAndUpdateStatus();
					if (updaterStarted == true) {
						for (int i = 0; i < (twitterUpdateInterval / 2); i++) {
							if (updaterStarted == true) Thread.sleep(1000);
						}
					}
					if (updaterStarted == true) {
						if (resetCounter == resetCounterMax) {
							resetCounter = 0;
							if (serviceAvailable == true) twitter.getUserTimeline();
						}
						else if (resetCounter < resetCounterMax) {
							resetCounter = resetCounter + 1;
						}
						else {
							resetCounter = 0;
						}
					}
				} 
				catch (InterruptedException e) {}
			}
			
			updater.interrupt();   
			System.out.println("Updating twitter status stopped!");
		}
		
		/**
	    * Generates the status and sends it to twitter
	    */
		private void generateAndUpdateStatus() {
			
			String status = "";
			String ret = null;

			try {
				Iterator<String> it2 = groupCaption.iterator();
				Iterator<ArrayList<String>> it3 = lineText.iterator();
				Iterator<ArrayList<String>> it4 = lineFormat.iterator();
				for (Iterator<ArrayList<String>> it1 = lineUri.iterator(); it1.hasNext();) {
					String caption = (String)it2.next(); 
					ArrayList<String> uriList = (ArrayList<String>)it1.next(); 
					ArrayList<String> textList = (ArrayList<String>)it3.next(); 
					ArrayList<String> formatList = (ArrayList<String>)it4.next(); 
					if (nextTweetIndex == groupCaption.indexOf(caption)) {
						status = caption + ": ";
						Iterator<String> it7 = textList.iterator();
						Iterator<String> it8 = formatList.iterator();
						for (Iterator<String> it6 = uriList.iterator(); it6.hasNext();) {
							String text = (String)it7.next(); 
							String format = (String)it8.next(); 
							ret = characterEscaper.decodeString((commonDeviceAccess.getMonitorData((String)(it6.next()))).split(">")[1].split("<")[0]);
							if (format.equals("") || format.equals(" ")) {  //process ret as string as it is
								status = status + text + ret + " ";
							}
							else if (format.contains("x")) {                //fixed-length string, process ret as string
								if (format.length() <= ret.length()) {
									status = status + text + ret.substring(0, format.length() - 1) + " ";
								}
								else {
									status = status + text + ret + " ";
								}
							}
							else {                                          //number format, process ret as number
								DecimalFormat formatter = new DecimalFormat(format.replace("n", "#"), new DecimalFormatSymbols(localSettings));
								status = status + text + formatter.format(Double.valueOf(ret)) + " ";
							}
						}
						status = status.substring(0, status.length() - 1);
					}
				}
			}
			catch (Exception e) {
				status = "Sorry! No status available.";
			}
					
			if (serviceAvailable == true) {
				twitter.updateStatus(status);
//				System.out.println("Last tweet: " + status);
				tweet = status;
			}
			
			if (nextTweetIndex < groupCaption.toArray().length - 1) {
				nextTweetIndex = nextTweetIndex + 1;
			}
			else {
				nextTweetIndex = 0;
			}
		}
	}
}