package de.desy.web2cToolkit.web2cUtil.web2cFileIO;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import de.desy.web2cToolkit.web2cUtil.web2cInterfaces.IWeb2cConstants;

import java.util.Iterator;

/**
 * @author R.Bacher, DESY
 */
public class Web2cLogbookRecordWriterAndReader implements IWeb2cConstants {
	
	private String logbookName = "";
	private String logbookPeriode = "";
	private String logbookDir = "";

	/**
	 * Writer and reader for Web2cLogbook files
	 * @param name logbook name
	 * @param periode logbook periode
	 * @param logbookHome directory which contains logbook-specific files
	 */
	public Web2cLogbookRecordWriterAndReader(String name, String periode, String home) {
		
		logbookName = name;
		logbookPeriode = periode;
		logbookDir = home;
	}
	
	/**
	 * Returns the first record in the logbook file specified
	 * @param fileName file name
	 * @param index record index
	 * @return record
	 */
	public String getFirstRecord(String fileName, int index) {

		return convertArrayListToString(readLogbookRecord(fileName, getFirstVisibleRecordIndex(fileName, index)));
	}	
	
	/**
	 * Returns the most recent record in the logbook file specified
	 * @param fileName file name
	 * @param index record index
	 * @return record
	 */
	public String getMostRecentRecord(String fileName, int index) {

		return convertArrayListToString(readLogbookRecord(fileName, getMostRecentVisibleRecordIndex(fileName, index)));
	}
	
	/**
	 * Returns next record in the logbook file specified
	 * @param fileName file name
	 * @param index record index
	 * @return record
	 */
	public String getNextRecord(String fileName, int index) {

		return convertArrayListToString(readLogbookRecord(fileName, getNextVisibleRecordIndex(fileName, index)));
	}
	
	/**
	 * Returns the previous record in the logbook file specified
	 * @param fileName file name
	 * @param index record index
	 * @return record
	 */
	public String getPreviousRecord(String fileName, int index) {

		return convertArrayListToString(readLogbookRecord(fileName, getPreviousVisibleRecordIndex(fileName, index)));
	}
	
	/**
	 * Returns the record in the logbook file specified
	 * @param fileName file name
	 * @param index record index
	 * @return record
	 */
	public String getRecord(String fileName, int index) {

		return convertArrayListToString(readLogbookRecord(fileName, index));
	}
	
	/**
	 * Returns the record text in the logbook file specified
	 * @param fileName file name
	 * @param index record index
	 * @return record text
	 */
	public String getRecordText(String fileName, int index) {

		return parseString(convertArrayListToString(readLogbookRecord(fileName, index)), "text");
	}
	
	/**
	 * Edits the record in the logbook file specified
	 * @param fileName file name
	 * @param index index of record to be edited
	 * @param recordString content of record to be saved
	 * @param authorName author name
	 * @return index of record edited
	 */
	public int editRecord(String fileName, int index, String recordString, String authorName) {
		
		return writeLogbookRecord(fileName, index, recordString, 1, authorName);
	}
	
	/**
	 * Appends the record to the most recent logbook file
	 * @param fileName file name
	 * @param recordString content of record to be saved
	 * @param authorName author name
	 * @return index of record appended
	 */
	public int appendRecord(String fileName, String recordString, String authorName) {
		
		return writeLogbookRecord(fileName, -1, recordString, 1, authorName);
	}
	
	/**
	 * Delets the record in the logbook file specified
	 * @param fileName file name
	 * @param index index of record to be deleted
	 * @return index of record deleted
	 */
	public int deleteRecord(String fileName, int index, String recordString) {
		
		return writeLogbookRecord(fileName, index, recordString, -1, "");
	}
	
	/**
	 * Returns the logbook tree
	 * @param listOrder 1: most recent file = last file; -1: most recent file = first file;
	 * @return tree
	 */
	public String getLogbookTree(int listOrder) {
		
		ArrayList<String> fileList = getLogbookFileList(listOrder);
		String str = "";
		for (Iterator<String> it = fileList.iterator(); it.hasNext();) {
			str = str + "<file>" + it.next() + "</file>";
		}
		
		return str;
	}
	
	/**
	 * Returns the most recent logbook gile
	 * @return file name
	 */
	public String getMostRecentLogbookFile() {
		
		return getLogbookFileList(-1).get(0);
	}
	
	/**
	 * Returns the logbook file list; creates a new file if the most recent file does not exist and adds it to the list
	 * @param listOrder 1: most recent file = last file; -1: most recent file = first file;
	 * @return list
	 */
	private ArrayList<String> getLogbookFileList(int listOrder) {
		
		ArrayList<String> fileList = new ArrayList<String>();
		boolean mostRecentFileExists = false;
		String mostRecentFileName = "";
		String shift = "";
		String day = "";
		String week = "";
		String month = "";
		String year = "";
		DecimalFormat formatter = null;
		
		Calendar calendar = Calendar.getInstance();
		int hour = calendar.get(Calendar.HOUR_OF_DAY);
		if (hour >= 7 && hour <= 14) shift = "M";
		else if (hour >= 15 && hour <= 23) shift = "E";
		else shift = "N";
		formatter = new DecimalFormat("00", new DecimalFormatSymbols(localSettings));
		day = formatter.format(calendar.get(Calendar.DAY_OF_MONTH));
		formatter = new DecimalFormat("00", new DecimalFormatSymbols(localSettings));
		week = "W" + formatter.format(calendar.get(Calendar.WEEK_OF_YEAR));
		formatter = new DecimalFormat("00", new DecimalFormatSymbols(localSettings));
		month = formatter.format(calendar.get(Calendar.MONTH) + 1);
		formatter = new DecimalFormat("0000", new DecimalFormatSymbols(localSettings));
		year =  formatter.format(calendar.get(Calendar.YEAR));	
		
		if (logbookPeriode.equals("shift")) {
			mostRecentFileName = logbookName + "_" + shift + "_" + day + "_" + month + "_" + year;
		}
		if (logbookPeriode.equals("day")) {
			mostRecentFileName = logbookName + "_" + day + "_" + month + "_" + year;
		}
		if (logbookPeriode.equals("week")) {
			mostRecentFileName = logbookName + "_" + week + "_" + year;
		}
		if (logbookPeriode.equals("month")) {
			mostRecentFileName = logbookName + "_" + month + "_" + year;
		}
		if (logbookPeriode.equals("year")) {
			mostRecentFileName = logbookName + "_" + year;
		}

		File dir = null;
		try {
			dir = (new File(logbookDir)).getCanonicalFile();
			String[] dirListing = dir.list();
			if (dirListing != null) {
				for (int i = 0; i < dirListing.length; i++) {
					if (dirListing[i].contains(logbookName)) {
						fileList.add(dirListing[i]);
						if (dirListing[i].equals(mostRecentFileName)) mostRecentFileExists = true;
					}
				}		
			}
			else {
				mostRecentFileExists = false;
			}
		} 
		catch (IOException e) {		
			System.out.println("Error browsing logbook directory: " + logbookDir);
			System.err.println(e.toString());
			e.printStackTrace();
		}
		
		if (mostRecentFileExists == false) {
			try {
				BufferedWriter out = new BufferedWriter(new FileWriter((new File(logbookDir + File.separator + mostRecentFileName)).getCanonicalFile()));
				out.write("<file>"); out.newLine();		
				out.write("</file>"); out.newLine();
				out.flush();
				out.close();
				fileList.add(mostRecentFileName);
			} 
			catch (IOException e) {
				System.out.println("Error writing logbook file: " + mostRecentFileName);			
				System.err.println(e.toString());
				e.printStackTrace();
			}
		}

		ArrayList<String> list = new ArrayList<String>();
		if (listOrder == -1) {
			for (int i = fileList.size() - 1; i >= 0; i--) {
				list.add(fileList.get(i));
			}
		}
		else {
			list = fileList;
		}
		
		return list;
	}
	
	/**
	 * Returns the number of records in the file specified
	 * @param fileName file name
	 * @return number of records
	 */
	private int getNumberOfRecords(String fileName) {
		
		int numberOfRecords = 0;
		List<Element> recordList = null;
		
		recordList = getRecords(fileName);
		if (recordList != null) numberOfRecords = recordList.size();
			
		return numberOfRecords;
	}
	
	/**
	 * Returns the records in the file specified
	 * @param fileName file name
	 * @return list of records
	 */
	@SuppressWarnings("unchecked")
	private List<Element> getRecords(String fileName) {
		
		List<Element> recordList = null;
		BufferedReader in = null;
		BufferedWriter out = null;

		try {
			in = new BufferedReader(new FileReader((new File(logbookDir + File.separator + fileName)).getCanonicalFile()));
			if (in.readLine() == null) {
				out = new BufferedWriter(new FileWriter((new File(logbookDir + File.separator + fileName)).getCanonicalFile()));
				out.write("<file>"); out.newLine();		
				out.write("</file>"); out.newLine();
				out.flush();
				out.close();
			}
			in.close();
		} 
		catch (FileNotFoundException e) {
			System.out.println("Error finding logbook file: " + fileName);
			System.err.println(e.toString());
			e.printStackTrace();
		} 
		catch (IOException e) {
			System.out.println("Error reading/writing logbook file: " + fileName);
			System.err.println(e.toString());
			e.printStackTrace();
		}
		
		try {		
		   	SAXBuilder builder = new SAXBuilder();
			Element root = builder.build(new File(logbookDir + File.separator + fileName).getCanonicalFile()).getRootElement();
			if (root.getName().equals("file")) recordList = (List<Element>)root.getChildren(); 
		}
		catch (JDOMException e) {
			System.out.println("Error parsing logbook file: " + fileName);
			System.err.println(e.toString());
			e.printStackTrace();
		} 
		catch (IOException e) {
			System.out.println("Error reading logbook file: " + fileName);
			System.err.println(e.toString());
			e.printStackTrace();
		}
			
		return recordList;
	}
	
	/**
	 * Returns the index of the visible record following the record with the index specified
	 * @param fileName file name
	 * @param index index of record
	 * @return number of records; if no record is found -1 is returned
	 */
	private int getNextVisibleRecordIndex(String fileName, int index) {
		
		int i = index + 1;
		int visibility = -1;
		ArrayList<String> list = new ArrayList<String>();
		
		if (checkForVisibleRecords(fileName) == false) {
			index = -1;
		}
		else {
			while ((visibility == -1) && (i < getNumberOfRecords(fileName))) {
				list = readLogbookRecord(fileName, i);
				for (Iterator<String> it = list.iterator(); it.hasNext();) {
					String element = it.next();
					if (element.contains("<visibility>")) visibility = Integer.valueOf(element.replace("<visibility>", "").replace("</visibility>", ""));
					if (element.contains("<index>") && (visibility == 1)) index = Integer.valueOf(element.replace("<index>", "").replace("</index>", ""));
				}
				i = i + 1;
			}
		}
		
		return index;
	}
	
	/**
	 * Returns the index of the visible record preceding the record with the index specified
	 * @param fileName file name
	 * @param index index of record
	 * @return number of records; if no record exists at all -1 is returned; if no record is found the index specified is returned
	 */
	private int getPreviousVisibleRecordIndex(String fileName, int index) {
		
		int i = index - 1;
		int visibility = -1;
		ArrayList<String> list = new ArrayList<String>();
		
		if (checkForVisibleRecords(fileName) == false) {
			index = -1;
		}
		else {
			while ((visibility == -1) && (i >= 0)) {
				list = readLogbookRecord(fileName, i);
				for (Iterator<String> it = list.iterator(); it.hasNext();) {
					String element = it.next();
					if (element.contains("<visibility>")) visibility = Integer.valueOf(element.replace("<visibility>", "").replace("</visibility>", ""));
					if (element.contains("<index>") && (visibility == 1)) index = Integer.valueOf(element.replace("<index>", "").replace("</index>", ""));
				}
				i = i - 1;
			}
		}
		
		return index;
	}
	
	/**
	 * Returns the index of the first visible record
	 * @param fileName file name
	 * @param index index of record
	 * @return number of records; if no record is found -1 is returned
	 */
	private int getFirstVisibleRecordIndex(String fileName, int index) {
		
		int i = 0;
		int visibility = -1;
		ArrayList<String> list = new ArrayList<String>();
		
		if (checkForVisibleRecords(fileName) == false) {
			index = -1;
		}
		else {
			while ((visibility == -1) && (i < getNumberOfRecords(fileName))) {
				list = readLogbookRecord(fileName, i);
				for (Iterator<String> it = list.iterator(); it.hasNext();) {
					String element = it.next();
					if (element.contains("<visibility>")) visibility = Integer.valueOf(element.replace("<visibility>", "").replace("</visibility>", ""));
					if (element.contains("<index>") && (visibility == 1)) index = Integer.valueOf(element.replace("<index>", "").replace("</index>", ""));
				}
				i = i + 1;
			}
		}
		
		return index;
	}
	
	/**
	 * Returns the index of the first visible record
	 * @param fileName file name
	 * @param index index of record
	 * @return number of records; if no record is found -1 is returned
	 */
	private int getMostRecentVisibleRecordIndex(String fileName, int index) {
		
		int i = getNumberOfRecords(fileName) - 1;
		int visibility = -1;
		ArrayList<String> list = new ArrayList<String>();
		
		if (checkForVisibleRecords(fileName) == false) {
			index = -1;
		}
		else {
			while ((visibility == -1) && (i >= 0)) {
				list = readLogbookRecord(fileName, i);	
				for (Iterator<String> it = list.iterator(); it.hasNext();) {
					String element = it.next();	
					if (element.contains("<visibility>")) visibility = Integer.valueOf(element.replace("<visibility>", "").replace("</visibility>", ""));
					if (element.contains("<index>") && (visibility == 1)) index = Integer.valueOf(element.replace("<index>", "").replace("</index>", ""));
				}
				i = i - 1;
			}
		}
		
		return index;
	}
	
	/**
	 * Checks for visible records
	 * @param fileName file name
	 * @return false = all records invisible; true = at least one record visible
	 */
	@SuppressWarnings("unchecked")
	private boolean checkForVisibleRecords(String fileName) {
		
		boolean visible = false;
		List<Element> recordList = null;
		
		recordList = getRecords(fileName);
		if (recordList.isEmpty() || recordList == null) visible = false;
		else {
			int i = 0; 
			for (Iterator<Element> it1 = recordList.iterator(); it1.hasNext();) {
				Element record = it1.next();
				List<Element> recordElementList = record.getChildren();
				for (Iterator<Element> it2 = recordElementList.iterator(); it2.hasNext();) {
					Element element = it2.next();	
					if (element.getName().equals("visibility") && (element.getValue().equals("1"))) i = i + 1;
				}
			}
			if (i > 0) visible = true;
		}
		
		return visible;
	}
	
    /**
     * Writes logbook record of the logbook file specified (modifies an existing record or appends a new record)
     * @param fileName logbook file name
     * @param index index of record <0 == record will be created; >=0 == record will be modified
     * @param record record content
     * @param visibility -1 == invisible; 1 == visible
     * @param authorName author name
     * @return index of record written (i < 0: error writing record)
     */
	private int writeLogbookRecord(String fileName, int index, String recordString, int visibility, String authorName) {
		
		int numberOfRecords = 0;		
		ArrayList<ArrayList<String>> records = new ArrayList<ArrayList<String>>();	
		
		Calendar calendar  = Calendar.getInstance();
		String now = (DateFormat.getDateTimeInstance(DateFormat.SHORT, DateFormat.SHORT, localSettings)).format(calendar.getTime());

		numberOfRecords = getNumberOfRecords(fileName);
		if (numberOfRecords > 0) for (int j = 0; j < numberOfRecords; j++) records.add(readLogbookRecord(fileName, j));
		
		if ((index == -1) || ((index >= 0) && (numberOfRecords > 0))) {
			try {
				BufferedWriter out = new BufferedWriter(new FileWriter(new File(logbookDir + File.separator + fileName).getCanonicalFile()));
				out.write("<file>"); out.newLine();	   	
				if (index < 0) {  //append record
					if (numberOfRecords > 0) {
						int i = 0;
						for (Iterator<ArrayList<String>> it1 = records.iterator(); it1.hasNext();) {
							out.write("    <record>"); out.newLine();	
							ArrayList<String> valueList = it1.next();
							for (Iterator<String> it2 = valueList.iterator(); it2.hasNext();) {
								out.write("        " + it2.next()); out.newLine();
							}
							out.write("    </record>"); out.newLine();
							i = i + 1;
						}
						index = i;
					}
					else {
						index = 0;
					}
					out.write("    <record>"); out.newLine();		
			    	out.write("        <visibility>" + visibility + "</visibility>"); out.newLine();
			    	out.write("        <index>" + index + "</index>"); out.newLine();
			    	out.write("        <created>" + now + "</created>"); out.newLine();
			    	out.write("        <modified>" + now + "</modified>"); out.newLine();
			    	out.write("        <author>" + authorName + "</author>"); out.newLine();
			    	out.write("        <caption>" + parseString(recordString, "caption") + "</caption>"); out.newLine();
			    	out.write("        <category>" + parseString(recordString, "category") + "</category>"); out.newLine();
			    	out.write("        <importance>" + parseString(recordString, "importance") + "</importance>"); out.newLine();
			    	out.write("        <text>" + parseString(recordString, "text") + "</text>"); out.newLine();
					out.write("    </record>"); out.newLine();
				}	
				else {            //modify record (i.e. edit or delete)
					if (index == 0) {
						out.write("    <record>"); out.newLine();		
				    	out.write("        <visibility>" + visibility + "</visibility>"); out.newLine();
			    		out.write("        <index>" + index + "</index>"); out.newLine();
				    	out.write("        <created>" + parseString(recordString, "created") + "</created>"); out.newLine();
				    	out.write("        <modified>" + now + "</modified>"); out.newLine();
			    		out.write("        <author>" + parseString(recordString, "author") + "</author>"); out.newLine();
			    		out.write("        <caption>" + parseString(recordString, "caption") + "</caption>"); out.newLine();
			    		out.write("        <category>" + parseString(recordString, "category") + "</category>"); out.newLine();
			    		out.write("        <importance>" + parseString(recordString, "importance") + "</importance>"); out.newLine();
			    		out.write("        <text>" + parseString(recordString, "text") + "</text>"); out.newLine();
						out.write("    </record>"); out.newLine();	
						int i = 0;
						for (Iterator<ArrayList<String>> it1 = records.iterator(); it1.hasNext();) {
							ArrayList<String> valueList = it1.next();
							if (i > index) {
								out.write("    <record>"); out.newLine();	
								for (Iterator<String> it2 = valueList.iterator(); it2.hasNext();) {
									out.write("        " + it2.next()); out.newLine();
								}
								out.write("    </record>"); out.newLine();
							}
							i = i + 1;
						}
					}
					else if ((index > 0) && (index < numberOfRecords - 1)) {
						int i = 0;
						for (Iterator<ArrayList<String>> it1 = records.iterator(); it1.hasNext();) {
							ArrayList<String> valueList = it1.next();
							if (i < index) {
								out.write("    <record>"); out.newLine();	
								for (Iterator<String> it2 = valueList.iterator(); it2.hasNext();) {
									out.write("        " + it2.next()); out.newLine();
								}
								out.write("    </record>"); out.newLine();
							}
							i = i + 1;
						}
						out.write("    <record>"); out.newLine();	
				    	out.write("        <visibility>" + visibility + "</visibility>"); out.newLine();
			    		out.write("        <index>" + index + "</index>"); out.newLine();
				    	out.write("        <created>" + parseString(recordString, "created") + "</created>"); out.newLine();
				    	out.write("        <modified>" + now + "</modified>"); out.newLine();
			    		out.write("        <author>" + parseString(recordString, "author") + "</author>"); out.newLine();
			    		out.write("        <caption>" + parseString(recordString, "caption") + "</caption>"); out.newLine();
			    		out.write("        <category>" + parseString(recordString, "category") + "</category>"); out.newLine();
			    		out.write("        <importance>" + parseString(recordString, "importance") + "</importance>"); out.newLine();
			    		out.write("        <text>" + parseString(recordString, "text") + "</text>"); out.newLine();
						out.write("    </record>"); out.newLine();	
						i = 0;
						for (Iterator<ArrayList<String>> it1 = records.iterator(); it1.hasNext();) {
							ArrayList<String> valueList = it1.next();
							if (i > index) {
								out.write("    <record>"); out.newLine();	
								for (Iterator<String> it2 = valueList.iterator(); it2.hasNext();) {
									out.write("        " + it2.next()); out.newLine();
								}
								out.write("    </record>"); out.newLine();
							}
							i = i + 1;
						}
					}
					else if (index == (numberOfRecords - 1)) {
						int i = 0;
						for (Iterator<ArrayList<String>> it1 = records.iterator(); it1.hasNext();) {
							ArrayList<String> valueList = it1.next();
							if (i < index) {
								out.write("    <record>"); out.newLine();	
								for (Iterator<String> it2 = valueList.iterator(); it2.hasNext();) {
									out.write("        " + it2.next()); out.newLine();
								}
								out.write("    </record>"); out.newLine();
							}
							i = i + 1;
						}
						out.write("    <record>"); out.newLine();		
				    	out.write("        <visibility>" + visibility + "</visibility>"); out.newLine();
			    		out.write("        <index>" + index + "</index>"); out.newLine();
				    	out.write("        <created>" + parseString(recordString, "created") + "</created>"); out.newLine();
				    	out.write("        <modified>" + now + "</modified>"); out.newLine();
			    		out.write("        <author>" + parseString(recordString, "author") + "</author>"); out.newLine();
			    		out.write("        <caption>" + parseString(recordString, "caption") + "</caption>"); out.newLine();
			    		out.write("        <category>" + parseString(recordString, "category") + "</category>"); out.newLine();
			    		out.write("        <importance>" + parseString(recordString, "importance") + "</importance>"); out.newLine();
			    		out.write("        <text>" + parseString(recordString, "text") + "</text>"); out.newLine();
						out.write("    </record>"); out.newLine();	
					}
				}			
				out.write("</file>"); out.newLine();
				out.flush();
				out.close();
			} 
			catch (IOException e) {
				System.out.println("Error writing logbook file: " + fileName); 
				System.err.println(e.toString());
				e.printStackTrace();
			}
		}
		
		return index;
	}
	
	/**
	 * Returns a record of the logbook file specified
	 * @param fileName logbook file name
	 * @param index index of record requested
	 * @return record 
	 */
	@SuppressWarnings("unchecked")
	private ArrayList<String> readLogbookRecord(String fileName, int index) {
		
		List<Element> recordList = null;
		ArrayList<String> valueList = new ArrayList<String>(); 
		
		recordList = getRecords(fileName);		
		if (recordList.isEmpty() || recordList == null || index < 0) {	
			valueList.add("<visibility>1</visibility>");
			valueList.add("<index>-1</index>");
			valueList.add("<created></created>");
			valueList.add("<modified></modified>");
			valueList.add("<author></author>");
			valueList.add("<caption></caption>");
			valueList.add("<category></category>");
			valueList.add("<importance></importance>");
			valueList.add("<text></text>");
		}
		else {
			if (index > recordList.size() - 1) index = recordList.size() - 1;
			int i = 0;
			for (Iterator<Element> it1 = recordList.iterator(); it1.hasNext();) {
				Element record = (Element)it1.next();
				if (i == index) {
					List<Element> recordElementList = (List<Element>)record.getChildren(); 
					for (Iterator<Element> it2 = recordElementList.iterator(); it2.hasNext();) {
						Element recordElement = (Element)it2.next();
						valueList.add("<" + recordElement.getName() + ">" + recordElement.getValue() + "</" + recordElement.getName() + ">");
					}
				}
				i = i + 1;
			}
		}
		
		return valueList;
	}
	
	/**
	 * Converts an ArrayList list to a String
	 * @return text string
	 */
	private String convertArrayListToString(ArrayList<String> list) {

		String ret = "<record>";

		for (Iterator<String> it = list.iterator(); it.hasNext();) {
			ret = ret + it.next();
		}
		
		ret = ret + "</record>";
		
		return ret;
	}
	
	/**
	 * Parses a string and finds the value associated with the key specified
	 * @param string string to be parsed
	 * @parameter key key string
	 * @return value
	 */
	private String parseString(String string, String key) {
		
		if (string.contains(key)) {
			String value = string.replace("<record>", "").replace("</record>", "").split("<" + key + ">")[1].split("</" + key + ">")[0];
			return value;
		}
		else {
			String value = "";
			return value;
		}
	}
}
