/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.displayers;

import java.beans.PropertyVetoException;
import java.util.Map;

import javax.swing.Action;

import com.cosylab.gui.InfoDialog;
import com.cosylab.gui.WheelswitchDisplayer;
import com.cosylab.gui.displayers.DataConsumer;
import com.cosylab.gui.displayers.DataState;
import com.cosylab.gui.displayers.DisplayerUtilities;
import com.cosylab.gui.displayers.DoubleConsumer;
import com.cosylab.gui.displayers.DoubleConsumerMulticaster;
import com.cosylab.gui.displayers.DoubleSeqConsumer;
import com.cosylab.gui.util.UserSettingsProtection;
import com.cosylab.util.CommonException;

import de.desy.acop.displayers.tools.AcopDisplayer;
import de.desy.acop.displayers.tools.AcopDisplayerTransferHandler;
import de.desy.acop.displayers.tools.AcopInfoDialog;
import de.desy.acop.transport.ConnectionParameters;
import de.desy.acop.transport.adapters.AcopTransportDataSource;
import de.desy.acop.transport.adapters.AdapterFactory;
import de.desy.acop.transport.adapters.AdapterFactoryService;
import de.desy.tine.definitions.TArrayType;
  
/**
 * <code>AcopWheelswitch</code> is a displayer which can present a single 
 * double value. The displayer consists of several digits and an up/down button.
 * In normal mode digits present the value, which is read from the remote 
 * connection point. When the user presses the up/down button the wheelswitch 
 * assumes the settable mode - in this case the user can enter the value he 
 * wants to set on the remote property.
 * <p>
 * For proper display of values one have to set the minimum and maximum values 
 * on the displayer. If the actual value is out of the prescribed bounds,
 * the displayer will notify the user by changing its visual appearance.
 * </p>
 * <p>
 * The connection point for this displayer should be set using the 
 * <i>connectionParameters</i> property, where the remote name of the 
 * {@link ConnectionParameters} points to the desired property.
 * </p>
 * 
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: Templates.xml,v 1.10 2004/01/13 16:17:13 jbobnar Exp $
 *
 */
public class AcopWheelswitch extends WheelswitchDisplayer implements  
	AcopDisplayer {
 
	private static final long serialVersionUID = -3337991208984786063L;
	private ConnectionParameters connectionParameters;
	private int arrayIndex=0;
	private AcopInfoDialog dialog;
	
	/**
	 * Constructs new AcopWheelswitch.
	 *
	 */
	public AcopWheelswitch() {
		super();
		setTiltingEnabled(true);
		new AcopDisplayerTransferHandler(this);
		UserSettingsProtection.setProtection(this,DisplayerUtilities.COMMON_NUMERIC_DISPLAYER_PROPERTIES,false);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DoubleSeqConsumer#updateValue(long, double[])
	 */
	public void updateValue(long timestamp, double[] value) throws CommonException {
	    updateValue(timestamp, DisplayerUtilities.extract(arrayIndex,value));	
    }

	/* (non-Javadoc)
	 * @see com.cosylab.gui.displayers.DataConsumer#getDataConsumer(java.lang.Class)
	 */
	@Override
	public DataConsumer getDataConsumer(Class type)
	{
		if (type == DoubleConsumer.class || type == DoubleSeqConsumer.class) {
			return this;
		}
		
		return DoubleConsumerMulticaster.createDataConsumer(type, this);
	}
	
	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.DialKnobDisplayer#setCharacteristics(java.util.Map)
	 */
	@Override
	public void setCharacteristics(Map characteristics) {
		super.setCharacteristics(characteristics);
		
		Object o  = characteristics.get("arrayType");
		if (o != null) {
			if (((TArrayType)o).isChannel()) {
				try {
					UserSettingsProtection.setUnprotected(this,"arrayIndex",0);
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		}
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopDisplayerConnector#getConnectionParameters()
	 */
	public ConnectionParameters getConnectionParameters() {
		return connectionParameters;
	}
	
	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.AcopDisplayerConnector#setConnectionParameters(de.desy.acop.transport.ConnectionParameters)
	 */
	public void setConnectionParameters(ConnectionParameters param) throws CommonException, PropertyVetoException {
		if (param!=null && connectionParameters != null) {
			if(param.equals(connectionParameters)) return;
		}
		updateDataState(new DataState(DataState.UNDEFINED));
			
		ConnectionParameters old = connectionParameters;
		this.connectionParameters = param;
		AdapterFactory factory = AdapterFactoryService.getInstance().getAdapterFactory();
		if (getDataSource() != null)
			factory.releaseDataSource((AcopTransportDataSource) getDataSource());
		if (param == null) {
			setDataSource(null);
			setTitle("No Connection");
		} else {
			if (param.getPropertySize() == ConnectionParameters.AUTO_PROPERTY_SIZE) {
				param = param.deriveWithPropertySize(1);
			}
			AcopTransportDataSource ds = factory.createDataSource(param);
			setDataSource(ds);
		}
		
		firePropertyChange(CONNECTION_PARAMETERS_PROPERTY,old, connectionParameters);
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.AcopDisplayer#getArrayIndex()
	 */
	public int getArrayIndex() {
		return arrayIndex;
	}

	/*
	 * (non-Javadoc)
	 * @see de.desy.acop.displayers.tools.AcopDisplayer#setArrayIndex(int)
	 */
	public void setArrayIndex(int arrayIndex) {
		if (this.arrayIndex == arrayIndex) return;
		int oldValue = this.arrayIndex;
		this.arrayIndex = arrayIndex;
		firePropertyChange("arrayIndex", oldValue, this.arrayIndex);
	}

	/*
	 * (non-Javadoc)
	 * @see com.cosylab.gui.WheelswitchDisplayer#getInfoDialog()
	 */
	@Override
	public InfoDialog getInfoDialog() {
		if (dialog == null) {
			dialog = new AcopInfoDialog(this);
		}
		return dialog;
	}
	
	/**
	 * Enables/disables Properties item in the popup menu.
	 * @param enable
	 */
	public void setPropertiesPopupEnabled(boolean enable) {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				a.setEnabled(enable);
				return;
			}
		}
	}
	
	/**
	 * Returns true if Properties item is enabled in the popup menu.
	 * @return
	 */
	public boolean isPropertiesPopupEnabled() {
		Action[] actions = getPopupManager().getActions();
		for (Action a : actions) {
			if (a != null && "Preferences...".equals(a.getValue(Action.NAME))) {
				return a.isEnabled();
			}
		}
		return false;
	}
}
/* __oOo__ */
