/*
 * Copyright (c) 2006 Stiftung Deutsches Elektronen-Synchroton,
 * Member of the Helmholtz Association, (DESY), HAMBURG, GERMANY.
 *
 * THIS SOFTWARE IS PROVIDED UNDER THIS LICENSE ON AN "../AS IS" BASIS.
 * WITHOUT WARRANTY OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR PARTICULAR PURPOSE AND
 * NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 * FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR
 * THE USE OR OTHER DEALINGS IN THE SOFTWARE. SHOULD THE SOFTWARE PROVE DEFECTIVE
 * IN ANY RESPECT, THE USER ASSUMES THE COST OF ANY NECESSARY SERVICING, REPAIR OR
 * CORRECTION. THIS DISCLAIMER OF WARRANTY CONSTITUTES AN ESSENTIAL PART OF THIS LICENSE.
 * NO USE OF ANY SOFTWARE IS AUTHORIZED HEREUNDER EXCEPT UNDER THIS DISCLAIMER.
 * DESY HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS,
 * OR MODIFICATIONS.
 * THE FULL LICENSE SPECIFYING FOR THE SOFTWARE THE REDISTRIBUTION, MODIFICATION,
 * USAGE AND OTHER RIGHTS AND OBLIGATIONS IS INCLUDED WITH THE DISTRIBUTION OF THIS
 * PROJECT IN THE FILE LICENSE.HTML. IF THE LICENSE IS NOT INCLUDED YOU MAY FIND A COPY
 * AT HTTP://WWW.DESY.DE/LEGAL/LICENSE.HTM
 */

package de.desy.acop.launcher;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Toolkit;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.lang.reflect.Constructor;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JPanel;

import com.cosylab.gui.components.util.ComponentPositioner;


/**
 * 
 * <code>Launcher</code> is a utility class which can launch an application
 * according to its type and supplied parameters.
 * 
 *
 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
 * @version $Id: Templates.xml,v 1.10 2004/01/13 16:17:13 jbobnar Exp $
 *
 */
public class Launcher {
	
	public static final String CAMNAME = "de.desy.mst.libs.framework.mstapp.util.camInstrumentation.CAMframeControl";
	
	/**
	 * 
	 * <code>StreamGobbler</code> reads the data from the input stream
	 * and forwards it to the outputstream.
	 *
	 * @author <a href="mailto:jaka.bobnar@cosylab.com">Jaka Bobnar</a>
	 *
	 */
	private static class StreamGobbler extends Thread {
		private InputStream is;
		private OutputStream os;


		StreamGobbler(InputStream is, OutputStream os) {
			this.is = is;
			this.os = os;
		}

		/**
		 * creates readers to handle the text created by the external program
		 */
		public void run() {
			try {
				BufferedReader reader = new BufferedReader(new InputStreamReader(is));
				PrintWriter writer = new PrintWriter(os);
				String line;
				while ((line = reader.readLine()) != null) {
					writer.println(line);
				}
				reader.close();
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
	}
	
	/**
	 * Method will start a webstart application. The provided String should be 
	 * the url to the jnlp file. This method will construct a webstart 
	 * start command composed of all provided parameters and forward the call to 
	 * {@link #runWebstartApplication(String, boolean)}.
	 * 
	 * @param jnlpURL the url of the jnlp file to be started
	 * @param jvmParameters the virtual machine parameters (without the -D)
	 * @param waitForCompletion the flag indicating if the calling thread
	 * 			should be blocked until the execution is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runWebstartApplication(String jnlpURL, String[] jvmParameters, boolean waitForCompletion) throws IOException {
		StringBuilder sb = new StringBuilder();
		for (String s : jvmParameters) {
			sb.append("-J-D");
			sb.append(s);
			sb.append(' ');
		}
		sb.append(jnlpURL);
		runWebstartApplication(sb.toString(), waitForCompletion);
	}
	
	/**
	 * Method will start a webstart application. The provided String should be 
	 * a complete url to the jnlp file. The provided url can also include 
	 * several program or JVM arguments. They have to be supplied as prescribed
	 * by the Java API. This method will forward the call to 
	 * {@link #runWebstartApplication(String, OutputStream, OutputStream, boolean)}
	 * where the streams are null.
	 *  
	 * @see #runCommand(String, OutputStream, OutputStream, boolean) 
	 *  
	 * @param completeJNLP_URL the url which will be loaded using the java
	 * 			webstart command (javaws)
	 * @param waitForCompletion the flag indicating if the calling thread
	 * 			should be blocked until the execution is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runWebstartApplication(String completeJNLP_URL, boolean waitForCompletion) throws IOException {
		runWebstartApplication(completeJNLP_URL, null, null, waitForCompletion);
	}

	/**
	 * Method will start a webstart application. The provided String should be 
	 * a complete url to the jnlp file. The provided url can also include 
	 * several program or JVM arguments. They have to be supplied as prescribed
	 * by the Java API.
	 *  
	 * @see #runCommand(String, OutputStream, OutputStream, boolean) 
	 *  
	 * @param completeJNLP_URL the url which will be loaded using the java
	 * 			webstart command (javaws)
	 * @param outputStream the output stream where all the output from the
	 * 			input stream of the executed process will be forwarded
	 * @param errorOutput the output stream where all the output from the
	 * 			error stream of the executed process will be forwarded
	 * @param waitForCompletion the flag indicating if the calling thread
	 * 			should be blocked until the execution is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runWebstartApplication(String completeJNLP_URL, OutputStream outputStream, OutputStream errorOutput, boolean waitForCompletion) throws IOException {
		runCommand("javaws " + completeJNLP_URL, outputStream, errorOutput, waitForCompletion);
	}
	
	/**
	 * Runs the specified command. This method will invoke the {@link Runtime#exec(String)},
	 * where the provided string parameters is the command. The output generated by
	 * the execution of this command is forwarded to the provided {@link OutputStream}s.
	 * If any of these stream are null the output from that part of execution will
	 * be skipped.
	 * <p>
	 * The invoker has the possibility to block the current thread until the 
	 * process created by this method is completed. The lifecycle of the process
	 * is specified by the command itself and the process which will be executed
	 * by that command.
	 * 
	 * @param command the command to be executed
	 * @param outputStream the output stream where all the data from the {@link Process}'s
	 * 			input stream will be forwarded (null is allowed)
	 * @param errorOutput the output stream where all the data from the {@link Process}'s
	 * 			error stream will be forwarded (null is allowed)
	 * @param waitForCompletion a flag indicating if the calling thread should be blocked
	 * 			until the process is completed
	 * @throws IOException if execution of the command failed
	 */
	public static void runCommand(String command, OutputStream outputStream, OutputStream errorOutput, boolean waitForCompletion) throws IOException {
		Process p = Runtime.getRuntime().exec(command);
		if (outputStream != null) {
			new StreamGobbler(p.getInputStream(), outputStream).start();
		}
		if (errorOutput != null) {
			new StreamGobbler(p.getErrorStream(), errorOutput).start();
		}
		if (waitForCompletion) {
			try {
				p.waitFor();
			} catch (InterruptedException e) {
				//ignore
			}
		}
	}
	
	
	private static String getClassName() {
		return Launcher.class.getName();
	}
	
	public static void main(String[] args) throws IOException {
		String[] s = new String[2];
		s[0] = "-Dtine.transport=TCP";
		s[1] = "-Dav.machine=FLASH";
		Launcher.runWebstartApplication("http://mcalaunch.desy.de/webapps/Released/common/ser/ArchiveViewer.jnlp", System.out, System.out, false);
	}
	
	/**
	 * Launches an application using the input arguments. This call is forwarded
	 * to {@link #launch(Class, String[], boolean)}, where shouldExitOnClose is false.
	 * 
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class decribing that object
	 * @param args program arguments
	 * @return the instance of the application
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args) {
		return launch(applicationClass, args, false);
	}
		
	/**
	 * Launches an application using the input arguments. This call is forwarded
	 * to {@link #launch(Class, String[], boolean)}, where shouldExitOnClose is given.
	 * 
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class decribing that object
	 * @param args program arguments
	 * @param shouldExitOnClose should the application be close on exit
	 * @return the instance of the application
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args, boolean shouldExitOnClose) {
		return launch(applicationClass, args, shouldExitOnClose, true);
	}
	/**
	 * Launches an application defined by the applicationClass. The method will try
	 * to create an object of the given instance using the constructor with String[]
	 * parameter. If this fails, it will call invoke the main method on the the 
	 * give application class using the args arguments. If the main method doesn't
	 * exist an object's default constructor will be used (in this case args array is
	 * useless). 
	 * <p>
	 * If the application is not a window type but it is at least a {@link Component},
	 * a JFrame will be created and the component will be added to that frame.
	 * DefaultCloseOperation for the frame is EXIT or DISPOSE and it depends
	 * on the shouldExitOnClose parameter. 
	 * </p>
	 * <p>
	 * This method is not thread safe and will block the current thread until the
	 * application is loaded. One should take of appropriate multithreading to avoid
	 * GUI freezing.
	 * </p>
	 *  
	 * @param <T> the object which should be constructed and returned
	 * @param applicationClass the class describing that object
	 * @param args the program arguments
	 * @param shouldExitOnClose if true the frame will have EXIT_ON_CLOSE defaultCloseOperation
	 * @param shouldDisposeOnClose if true the frame will have DISPOSE_ON_CLOSE onless exit is requested
	 * @return the instance of application if it was started
	 */
	public static <T> T launch(final Class<T> applicationClass, final String[] args, boolean shouldExitOnClose, boolean shouldDisposeOnClose) {
		if (applicationClass == null) {
			System.out.println("No class specified to launch in '"+getClassName()+".launch()' method");
			return null;
		}
		if (args == null) {
			System.out.println("No arguments specified in '"+getClassName()+".launch()' method.");
			return null;
		}
				
		Constructor<T> constructor = null;
		T obj;
		try {
			constructor = applicationClass.getConstructor(new Class[]{String[].class});
		} catch (NoSuchMethodException e2) {
			//ignore
		}
		
		if (constructor == null) {
			try {
				obj = applicationClass.newInstance();
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println("Cannot launch class '"+applicationClass + "'. \nIt does not provide a Constructor with parameters 'String[]' nor default constructor.");
				return null;
			}

			/*
			 * This code below effecitvly prevents you to launch class from 
			 * main method, it makes closed loop.
			 * This happens only if you launch it from the main method that invoked
			 * this Launcher call.
			 */
			
			/*//if no String[] constructor try main method
			Method mainMethod = null;
			try {
				mainMethod = applicationClass.getMethod("main", new Class[]{String[].class});
			} catch (NoSuchMethodException e1) {
				//ignore
			}

			if (mainMethod != null) {
				try {
					mainMethod.invoke(null, new Object[]{args});
					return null;
				} catch (Exception e) {
					//ignore
				}
			}
			//if no main method try default constructor
			try {
				constructor = applicationClass.getConstructor(new Class[]{});
			} catch (NoSuchMethodException e2) {
				//ignore
			}
			//if no default constructor abort
			if (constructor == null) {
				System.out.println("Cannot launch class '"+applicationClass + "'. \nIt does not provide a Constructor with parameters 'String[]' nor default constructo or the main method.");
				return null;
			}
			
			//create new object
			try {
				obj = constructor.newInstance(new Object[]{});
			} catch (Exception e) {
				e.printStackTrace();
				return null;
			}*/
			
		} else {
			
			try {
				obj = constructor.newInstance(new Object[]{args});
			} catch (Exception e) {
				e.printStackTrace();
				return null;
			}
		}
		if (obj != null) {
			int closeOperation = shouldExitOnClose ? JFrame.EXIT_ON_CLOSE : shouldDisposeOnClose ? JFrame.DISPOSE_ON_CLOSE : JFrame.DO_NOTHING_ON_CLOSE;
			if (JFrame.class.isAssignableFrom(applicationClass)) {
				JFrame frame = JFrame.class.cast(obj);
				frame.setDefaultCloseOperation(closeOperation);
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
				initializeFrameControl(frame);
			} else if (JDialog.class.isAssignableFrom(applicationClass)) {
				JDialog frame = JDialog.class.cast(obj);
				frame.setDefaultCloseOperation(closeOperation);
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
			} else if (JInternalFrame.class.isAssignableFrom(applicationClass)) { 
				JInternalFrame frame = JInternalFrame.class.cast(obj);
				frame.setDefaultCloseOperation(closeOperation);
				frame.setVisible(true);
			} else if (Container.class.isAssignableFrom(applicationClass)) {
				Container comp = Container.class.cast(obj);
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(closeOperation);
				frame.setContentPane(comp);
				Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setSize((int)(d.width*0.8), (int)(d.height*0.8));
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
				initializeFrameControl(frame);
			} else if (Component.class.isAssignableFrom(applicationClass)) {
				Component comp = Component.class.cast(obj);
				JFrame frame = new JFrame();
				frame.setDefaultCloseOperation(closeOperation);
				JPanel p = new JPanel(new BorderLayout());
				p.add(comp, BorderLayout.CENTER);
				frame.setContentPane(p);
				Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
				frame.setSize((int)(d.width*0.8), (int)(d.height*0.8));
				ComponentPositioner.centerOnScreen(frame);
				frame.setVisible(true);
				initializeFrameControl(frame);
			}
			return obj;
		}
		return null;
	}
	
	/**
	 * Initializes Component Application Manager. This method uses reflection to initialize
	 * the class, and than instantiates the manager.
	 * 
	 * @param frame the frame to connect it with the manager
	 * @see de.desy.mst.libs.framework.mstapp.util.camInstrumentation.CAMframeControl
	 */
	private static final void initializeFrameControl(Frame frame) {
		try {
    		Class<?> c = Class.forName(CAMNAME);
    		Constructor<?> constructor = c.getConstructor(new Class[]{Frame.class});
    		constructor.newInstance(frame);
		} catch (Exception e) {
			System.err.println("Cannot initialize CAM frame control.");
		}
	}
}
