/**
 * 
 */
package de.desy.acop.video;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.geom.GeneralPath;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import javax.swing.JViewport;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import de.desy.acop.video.analysis.AImage;
import de.desy.acop.video.analysis.ImageAnalysisEngine;
import de.desy.acop.video.displayer.ImageDisplayer;
import de.desy.acop.video.displayer.ImageZoom;
import de.desy.tine.types.IMAGE;
import de.desy.tine.types.IMAGE.FrameHeader;

/**
 * </code>DisplayerManager</code> manages the </code>ImageDisplayer</code> and the two
 * </code>SideProfileDisplayer</code>s contained in </code>AcopVideo</code>. It is
 * supposed to give a better overview of updating the displayers.
 * 
 * @author Tilen Kusterle, Cosylab
 *
 */
public class DisplayerManager extends MouseAdapter {
	
	public static final String PROPERTY_ON_SCREEN_ROI = "onScreenRoi";
	
	private final ImageAnalysisEngine analysisEngine;
	private final ImageDisplayer imageDisp;
	private final SideProfileDisplayer horizontalDisp;
	private final SideProfileDisplayer verticalDisp;
	
	private PropertyChangeSupport pcSupport;
	
	private ScaleHelper scaleHelper;
	private ChangeListener viewPortListener;
	private AImage aImage;
	
	private Rectangle2D selectedRectangle;
	private Rectangle2D thresholdRectangle;
	private Rectangle2D draggingRectangle;
	private Rectangle2D crossRect;
	private Rectangle2D crossFitRect;
	private Rectangle aoiRectangle;
	
	private boolean dragStarted = false;
	private Point2D p1 = null;
	private Point2D p2 = null;
	private Point2D viewPoint = new Point2D.Double();
	
	private boolean canChangeSettings = false;
	
	private boolean displayStatistics = true;
	
	public DisplayerManager(ImageAnalysisEngine imageAnalysisEngine, ImageDisplayer imageDisplayer, SideProfileDisplayer horizontalDisplayer, SideProfileDisplayer verticalDisplayer) {
		analysisEngine = imageAnalysisEngine;
		imageDisp = imageDisplayer;
		horizontalDisp = horizontalDisplayer;
		verticalDisp = verticalDisplayer;
		
		imageDisp.addMouseListener(this);
		imageDisp.addMouseMotionListener(this);
		imageDisp.addComponentListener(new ComponentAdapter() {
			@Override
			public void componentResized(ComponentEvent e) {
				getScaleHelper().setComponentDimesnion(imageDisp.getSize());
			}
		});
		imageDisp.addPropertyChangeListener(new PropertyChangeListener() {
			@Override
			public void propertyChange(PropertyChangeEvent evt) {
				String name = evt.getPropertyName();
				if (name == ImageDisplayer.PROPERTY_KEEP_ASPECT_RATIO) {
					getScaleHelper().setKeepAspectRatio((Boolean) evt.getNewValue());
				}
				else if (name == ImageDisplayer.PROPERTY_AOI_ZOOM) {
					getScaleHelper().setAoiZoom((Boolean) evt.getNewValue());
				}
				else if (name == ImageDisplayer.PROPERTY_IMAGE_ZOOM) {
					getScaleHelper().setImageZoom((ImageZoom) evt.getNewValue());
				}
			}
		});
	}
	
	public void setDisplayStatistics(boolean display) {
		this.displayStatistics = display;
		horizontalDisp.setDisplayStatistics(display);
		verticalDisp.setDisplayStatistics(display);
	}
	
	public boolean isDisplayStatistics() {
		return this.displayStatistics;
	}
	
	public void updateImageDisplayerSize(Dimension size) {
		getScaleHelper().setComponentDimesnion(size);
	}
	
	public synchronized void updateAImage(AImage aImage) {
		if (this.aImage == aImage) return;
		boolean clean = true;
		if (this.aImage != null && aImage != null) {
			clean = this.aImage.getImageW() != aImage.getImageW() ||
					this.aImage.getImageH() != aImage.getImageH();
		}
		this.aImage = aImage;
		if (dragStarted && clean) {
			p1 = null;
			p2 = null;
			dragStarted = false;
			draggingRectangle = null;
			horizontalDisp.clearRoi();
			verticalDisp.clearRoi();
		}
		if (aImage == null) {
			crossRect = null;
			selectedRectangle = null;
			getScaleHelper().setImageData(null, null);
			imageDisp.resetForReceiving();
			horizontalDisp.clearData();
			verticalDisp.clearData();
			return;
		}
		
		IMAGE im = aImage.getImage();
		FrameHeader fh = im.getFrameHeader();
		int offsetX = fh.xStart;
		int offsetY = fh.yStart;
		int sizeX = (fh.aoiWidth >= 0) ? fh.aoiWidth : fh.sourceWidth;
		int sizeY = (fh.aoiHeight >= 0) ? fh.aoiHeight : fh.sourceHeight;
//		getScaleHelper().setImageData(new Dimension(fh.sourceWidth, fh.sourceHeight), new Rectangle(offsetX, offsetY, sizeX, sizeY));
		aoiRectangle = new Rectangle(offsetX, offsetY, sizeX, sizeY);
		getScaleHelper().setImageData(new Dimension(fh.sourceWidth, fh.sourceHeight), aoiRectangle);
		
		crossRect = new Rectangle2D.Double(aImage.getMeanX()-aImage.getStdX()/2, aImage.getMeanY()-aImage.getStdY()/2, aImage.getStdX(), aImage.getStdY());
		
		selectedRectangle = new Rectangle2D.Double(aImage.getRoiX(), aImage.getRoiY(), aImage.getRoiW(), aImage.getRoiH());
		if (aImage.isCalculateThreshold()) {
			thresholdRectangle = new Rectangle2D.Double(aImage.getRoi2X(), aImage.getRoi2Y(), aImage.getRoi2W(), aImage.getRoi2H());
		} else {
			thresholdRectangle = null;
		}
		
		if (aImage.isPerformFit()) {
			crossFitRect = new Rectangle2D.Double(aImage.getMeanXFit() - aImage.getStdXFit()/2, aImage.getMeanYFit()-aImage.getStdYFit()/2,aImage.getStdXFit(),aImage.getStdYFit());
		} else {
			crossFitRect = null;
		}
		
		imageDisp.updateValue(im);
		// update decorations...
		adjustSideProfileAspectRatio(false);
		horizontalDisp.setData(
				aImage.getSideViewX(), aImage.getSideViewAmplitudeX(),
				aImage.getMeanX(), aImage.getStdX(), aImage.getSideViewConstX(), 
				aImage.getAmplitudeXFit(), aImage.getMeanXFit(), aImage.getStdXFit(), aImage.getConstXFit(),
				aImage.getSlopeXFit(),aImage.getRoiX(), aImage.getRoiW(), aImage.isPerformFit()
		);
		verticalDisp.setData(
				aImage.getSideViewY(), aImage.getSideViewAmplitudeY(),
				aImage.getMeanY(), aImage.getStdY(), aImage.getSideViewConstY(), 
				aImage.getAmplitudeYFit(), aImage.getMeanYFit(), aImage.getStdYFit(), aImage.getConstYFit(),
				aImage.getSlopeYFit(),aImage.getRoiY(), aImage.getRoiH(), aImage.isPerformFit()
		);
	}
	
	@Override
	public void mouseClicked(MouseEvent e) {
		if (e.getClickCount() == 2 && aImage != null) {
			Rectangle r = getScaleHelper().getImageAoi();
			if (canChangeSettings) analysisEngine.setRoi(0, 0, r.width, r.height);
			else getPcSupport().firePropertyChange(PROPERTY_ON_SCREEN_ROI, null, r);
		}
	}

	@Override
	public void mouseDragged(MouseEvent e) {
		if (dragStarted) {
			p2 = transformPoint(e.getPoint(), e.getComponent().getSize());
			horizontalDisp.setRoi(p1.getX()-viewPoint.getX(), p2.getX()-viewPoint.getX());
			verticalDisp.setRoi(p1.getY()-viewPoint.getY(), p2.getY()-viewPoint.getY());
			
			draggingRectangle = calculateRectangle();
			imageDisp.repaint();
		}
	}

	@Override
	public void mousePressed(MouseEvent e) {
		if (aImage == null) return;
		if (e.getButton() == MouseEvent.BUTTON1) {
			dragStarted = true;
			p1 = transformPoint(e.getPoint(), e.getComponent().getSize());
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if (e.getButton() == MouseEvent.BUTTON1) {
			if (dragStarted) {
				if (p1 != null && p2 != null) {
					
					Rectangle r = getScaleHelper().componentToImage(calculateRectangle());
					Rectangle rect = getScaleHelper().getImageAoi();
					r = new Rectangle(r.x + rect.x,r.y + rect.y,r.width,r.height);
					if (canChangeSettings) {
						horizontalDisp.clearRoi();
						verticalDisp.clearRoi();
						analysisEngine.setRoi(r.x, r.y, r.width, r.height);
					}
					else getPcSupport().firePropertyChange(PROPERTY_ON_SCREEN_ROI, null, r);
				}
				dragStarted = false;
				p1 = null;
				p2 = null;
				draggingRectangle = null;
			}
		}
	}
	
	private Rectangle2D calculateRectangle() {
		double x, y, w, h;
		if (p1.getX() < p2.getX()) {
			x = p1.getX();
			w = p2.getX()-p1.getX();
		} else {
			x = p2.getX();
			w = p1.getX()-p2.getX();
		}
		if (p1.getY() < p2.getY()) {
			y = p1.getY();
			h = p2.getY()-p1.getY();
		} else {
			y = p2.getY();
			h = p1.getY()-p2.getY();
		}
		return new Rectangle2D.Double(x, y, w, h);
	}
	
	private synchronized Point2D transformPoint(Point p, Dimension dim) {
		Rectangle2D r = getScaleHelper().getActiveRectangle();
		r = getScaleHelper().imageToComponent(getScaleHelper().componentToImage(r));
		double x = p.x, y = p.y;
		if (p.x < r.getMinX()) x = r.getMinX();
		else if (p.x > r.getMaxX()) x = r.getMaxX();
		if (p.y < r.getMinY()) y = r.getMinY();
		else if (p.y > r.getMaxY()) y = r.getMaxY();
		p.setLocation(x, y);
		return p;
	}
	
	public Rectangle2D getDraggingRectangle() {
		return draggingRectangle;
	}
	
	public boolean isRoiSelected() {
		return selectedRectangle != null;
	}
	
	public synchronized Rectangle2D getScaledSelectedRectangle() {
		if (selectedRectangle == null) return null;
		return getScaleHelper().imageToComponent(selectedRectangle);
	}
	
	public boolean isThresholdROISelected() {
		return thresholdRectangle != null;
	}
	
	public synchronized Rectangle2D getScaledThresholdROIRectangle() {
		if (thresholdRectangle == null) return null;
		return getScaleHelper().imageToComponent(thresholdRectangle);
	}
	
	public boolean isCrossAvailable() {
		return crossRect != null && displayStatistics;
	}
	
	public synchronized Shape getCross() {
		if (crossRect == null) return null;
		Rectangle2D r = getScaleHelper().imageToComponent(crossRect);
		if (r == null) return null;
		GeneralPath marker = new GeneralPath(new Line2D.Double(r.getCenterX()-r.getWidth()/2, r.getCenterY(), r.getCenterX()+r.getWidth()/2, r.getCenterY()));
		marker.append(new Line2D.Double(r.getCenterX(), r.getCenterY()-r.getHeight()/2, r.getCenterX(), r.getCenterY()+r.getHeight()/2), false);
		return marker;
	}
	
	public boolean isCrossFitAvailable() {
		return crossFitRect != null;
	}
	
	public synchronized Shape getCrossFit() {
		if (crossFitRect == null) return null;
		Rectangle2D r = getScaleHelper().imageToComponent(crossFitRect);
		if (r == null) return null;
		GeneralPath marker = new GeneralPath(new Line2D.Double(r.getCenterX()-r.getWidth()/2, r.getCenterY(), r.getCenterX()+r.getWidth()/2, r.getCenterY()));
		marker.append(new Line2D.Double(r.getCenterX(), r.getCenterY()-r.getHeight()/2, r.getCenterX(), r.getCenterY()+r.getHeight()/2), false);
		return marker;
	}
	
	public ChangeListener getViewPortListener() {
		if (viewPortListener == null) {
			viewPortListener = new ChangeListener() {
				@Override
				public void stateChanged(ChangeEvent e) {
					JViewport port = (JViewport) e.getSource();
					Point p = port.getViewPosition();
					if (p.getX() != viewPoint.getX() || p.getY() != viewPoint.getY()) {
						viewPoint = p;
						horizontalDisp.setCanvasStart(-viewPoint.getX());
						verticalDisp.setCanvasStart(-viewPoint.getY());
					}
				}
			};
			
		}

		return viewPortListener;
	}
	
	private synchronized ScaleHelper getScaleHelper() {
		if (scaleHelper == null) {
			scaleHelper = new ScaleHelper(imageDisp.isKeepAspectRatio(), imageDisp.isAOIZoom(), imageDisp.getImageZoom());
			scaleHelper.addPropertyChangeListener(new PropertyChangeListener() {
				@Override
				public void propertyChange(PropertyChangeEvent evt) {
					String name = evt.getPropertyName();
					if (name == ScaleHelper.PROPERTY_COMPONENT_DIMENSION) {
						adjustSideProfileAspectRatio(true);
					}
					else if (name == ScaleHelper.PROPERTY_ACTIVE_RECTANGLE) {
						adjustSideProfileAspectRatio(true);
					}
				}
			});
		}
		return scaleHelper;
	}
	
	private void adjustSideProfileAspectRatio(boolean repaintAtOnce) {
		Rectangle2D image = getScaleHelper().getActiveRectangle();
		if (image == null) return;
		image = getScaleHelper().imageToComponent(getScaleHelper().componentToImage(image));
		
		//+2 because borders are used
		horizontalDisp.setCanvasSize(image.getX()-viewPoint.getX(), image.getWidth()+2, repaintAtOnce);
		verticalDisp.setCanvasSize(image.getY()-viewPoint.getY(), image.getHeight()+2, repaintAtOnce);
	}

	/**
	 * @return the canChangeSettings
	 */
	public boolean isCanChangeSettings() {
		return canChangeSettings;
	}

	/**
	 * @param canChangeSettings the canChangeSettings to set
	 */
	public void setCanChangeSettings(boolean canChangeSettings) {
		this.canChangeSettings = canChangeSettings;
	}

	private PropertyChangeSupport getPcSupport() {
		if (pcSupport == null) {
			pcSupport = new PropertyChangeSupport(this);
		}
		return pcSupport;
	}

	/**
	 * Adds a </code>PropertyChangeListener</code> for the specified propertyName.
	 * @param propertyName the name of the property
	 * @param listener the </code>PropertyChangeListener</code> to add
	 */
	public void addPropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		getPcSupport().addPropertyChangeListener(propertyName, listener);
	}
	
	/**
	 * Removes a </code>PropertyChangeListener</code> for the specified propertyName.
	 * @param propertyName the name of the property
	 * @param listener the </code>PropertyChangeListener</code> to remove
	 */
	public void removePropertyChangeListener(String propertyName, PropertyChangeListener listener) {
		getPcSupport().removePropertyChangeListener(propertyName, listener);
	}
	
	/**
	 * Adds a </code>PropertyChangeListener</code>.
	 * @param listener the </code>PropertyChangeListener</code> to add
	 */
	public void addPropertyChangeListener(PropertyChangeListener listener) {
		getPcSupport().addPropertyChangeListener(listener);
	}
	
	/**
	 * Removes a </code>PropertyChangeListener</code>.
	 * @param listener the </code>PropertyChangeListener</code> to remove
	 */
	public void removePropertyChangeListener(PropertyChangeListener listener) {
		getPcSupport().removePropertyChangeListener(listener);
	}
	
}
